"""
"""

"""Enum for the different types of modulator the Modulator component can represent.

am = Amplitude modulation
pm = Phase modulation
"""
cpdef enum ModulatorType:
    am
    pm
    # TODO ddb need to add in alignment and mode-matching modulator



cdef extern from "math.h" nogil:
    double atan2(double y, double x)
    double cos(double arg)
    double exp(double arg)
    double fabs(double arg)
    double fmax(double x, double y)
    double log(double x)
    double sin(double arg)
    double sqrt(double arg)
    double asin(double arg)
    double acos(double arg)
    double cosh(double arg)
    double sinh(double arg)
    double ceil(double arg)
    double floor(double arg)
    int isfinite(float arg)
    float NAN
    float INF

cdef extern from "constants.h":
    long double PI
    double DEG2RAD
    double RAD2DEG

## TODO (sjr) Used the fused type below on some functions
##            in this file for template-like behaviour

#ctypedef fused real_numeric_t: # non-complex numeric type
#    short
#    int
#    long
#    float
#    double

cpdef inline double degrees(double x) nogil:
    return x * RAD2DEG


cpdef inline double radians(double x) nogil:
    return x * DEG2RAD

cpdef inline int sgn(double x) nogil:
    return (0.0 < x) - (x < 0.0)

cpdef inline double msign(int n) nogil:
    return -1.0 if n % 2 else 1.0


cpdef inline int nmin(int n, int m) nogil:
    return n if n < m else m


cpdef inline int nmax(int n, int m) nogil:
    return n if n > m else m


cpdef inline bint float_eq(double x, double y) nogil:
    if x == 0.0 and y == 0.0:
        return 1
    else:
        return fabs(x - y) / fmax(fabs(x), fabs(y)) < 1e-13


cpdef double factorial(int n) nogil
cpdef double sqrt_factorial(int n) nogil

cpdef double hermite(int n, double x) nogil



cimport cython

import numpy as np
cimport numpy as np
import logging

from libc.stdlib cimport free, malloc, calloc

from finesse.knm cimport (
    KnmMatrix,
    knm_bh_workspace,
    knm_bh_ws_init,
    knm_bh_ws_free,
    knm_bh_ws_is_changing,
    knm_bh_ws_recompute_mismatch,
    knm_bh_ws_recompute,
    knm_bh_ws_recompute_misalignment,
)
from finesse.knm.bayerhelms cimport (
    fast_compute_refl_knm_matrix_bh__k00_zeroed,
    fast_compute_trns_knm_matrix_bh__k00_zeroed,
    fast_compute_refl_knm_matrix_bh,
    fast_compute_trns_knm_matrix_bh,
)
from finesse.knm.matrix cimport knm_loss
from finesse.cmatrix cimport SubCCSView
from finesse.components.workspace cimport ConnectorWorkspace, FillFuncWrapper, NoiseInfo
from finesse.components.general import NodeType, NoiseType
from finesse.cymath.gaussbeam cimport c_transform_q
from finesse.frequency cimport frequency_info_t
from finesse.simulations.base cimport NodeBeamParam

import finesse.components as components


LOGGER = logging.getLogger(__name__)

cdef double[:,::1] abcd_unity = np.eye(2, dtype=float)


cdef class KnmConnectorWorkspace(ConnectorWorkspace):
    def __init__(self, object owner, CarrierSignalMatrixSimulation sim, *args, **kwargs):
        ConnectorWorkspace.__init__(self, owner, sim, *args, **kwargs)

        # Here we automatically generate the coupling strings used
        # for generating the Knm matricies. We use the optical to optical
        # connections that have been registered, the port number is taken
        # from the order in which they are added to the component. I guess
        # there could also be a more explicit defition of the port "index"
        # given in the _add_port method but we'll use this for now.

        # TODO ddb could store index in port object perhaps, along with the
        # list of optical ports
        self.o2o = owner.all_optical_connections
        oports = list(p for p in owner.ports if p.type == NodeType.OPTICAL)

        self.N_opt_ports = len(oports)
        self.N_opt_conns = len(self.o2o)

        self.onode_ids = <Py_ssize_t*> malloc(sizeof(Py_ssize_t) * self.N_opt_ports * 2)
        if not self.onode_ids:
            raise MemoryError()

        self.oconn_info = <KnmInfo*> calloc(sizeof(KnmInfo), self.N_opt_conns)
        if not self.oconn_info:
            raise MemoryError()

        # 2 lots per connection because of x and y planes
        self.Kws = <knm_bh_workspace*> calloc(sizeof(knm_bh_workspace), 2*self.N_opt_conns)
        if not self.Kws:
            raise MemoryError()

        # need these node ids for the simulation for indexing traces
        for i, p in enumerate(oports):
            self.onode_ids[2*i] = sim.carrier.node_id(p.i)
            self.onode_ids[2*i+1] = sim.carrier.node_id(p.o)

        cdef np.ndarray[double, ndim=1, mode='c'] scatter_loss
        for i, (conn, (f, t)) in enumerate(self.o2o.items()):
            # TODO ddb should probably use some fixed index rather than a list of the order
            # they are defined in the element definition
            a = oports.index(f.port)
            b = oports.index(t.port)
            coupling = f"{a+1}{b+1}"
            self.oconn_info[i].from_port_idx = a
            self.oconn_info[i].to_port_idx = b
            self.oconn_info[i].K_ws_x = &self.Kws[2*i]
            self.oconn_info[i].K_ws_y = &self.Kws[2*i+1]
            self.oconn_info[i].abcd_x = &abcd_unity[0,0]
            self.oconn_info[i].abcd_y = &abcd_unity[0,0]

            knm = KnmMatrix(self.sim.model_data.homs_view, self.owner.name, coupling)
            self.oconn_info[i].mtx = &knm.mtx
            # Create the actual Knm matrix here based on the optical
            # port coupling indices
            setattr(self, f"K{coupling}", knm)

            # Make the buffer of knm loss data
            scatter_loss = np.zeros(self.sim.model_data.num_HOMs)
            setattr(self, f"K{coupling}_loss", scatter_loss)
            # Keep ptr to the buffer for use in compute_knm_losses
            self.oconn_info[i].loss = &scatter_loss[0]

        self.total_losses = np.zeros(self.sim.model_data.num_HOMs)
        if sim.signal:
            self.signal.set_fill_noise_function(NoiseType.QUANTUM, optical_quantum_noise_knm)

    def __dealloc__(self):
        if self.onode_ids:
            free(self.onode_ids)
        if self.oconn_info:
            free(self.oconn_info)
        if self.Kws:
            free(self.Kws)

    cpdef set_knm_info(self, connection,
        double[:,::1] abcd_x = abcd_unity,
        double[:,::1] abcd_y = abcd_unity,
        double nr_from=1, double nr_to=1,
        bint is_transmission=False,
        Parameter beta_x = None,
        double beta_x_factor = 0,
        Parameter beta_y = None,
        double beta_y_factor = 0,
        Parameter alpha = None
    ):
        """Python facing method to set how Knm calculations should be handled for a
        particular connection. When the workspace is being created this should be called if modal
        simualtions are being used. This then sets which ABCD matricies to use, various parameters
        required, and whether the connection is a transmission or reflection.
        """
        cdef:
            KnmInfo *conn
            int index

        if abcd_x.shape[0] != abcd_x.shape[1] != 2:
            raise Exception("ABCD X is not 2x2")
        if abcd_y.shape[0] != abcd_y.shape[1] != 2:
            raise Exception("ABCD Y is not 2x2")
        # First find the connection str in our connections then store the information)
        if connection not in self.o2o:
            raise Exception(f"Connection {connection} is not a valid connection in the element {self.owner}")

        index = list(self.o2o.keys()).index(connection)
        conn = &self.oconn_info[index]
        conn.has_been_set = True
        conn.abcd_x = &abcd_x[0, 0]
        conn.abcd_y = &abcd_y[0, 0]
        conn.is_transmission = is_transmission
        conn.nr_from = nr_from
        conn.nr_to = nr_to
        if beta_x:
            conn.beta_x = &beta_x.__cvalue
            conn.beta_x_is_changing = beta_x.is_changing if beta_x else False
            conn.beta_x_factor = beta_x_factor
        if beta_y:
            conn.beta_y = &beta_y.__cvalue
            conn.beta_y_is_changing = beta_y.is_changing if beta_y else False
            conn.beta_y_factor = beta_y_factor
        if alpha:
            conn.alpha = &alpha.__cvalue
            conn.alpha_is_changing = alpha.is_changing
            if alpha.is_changing:
                raise NotImplementedError("Changing angle of incidence not supported yet")

        # Set the appropriate scatter matrix function to use based on
        # whether this is a transmission or reflection coupling, and
        # also whether the 00->00 coupling coefficient should have
        # zeroed phase (i.e. zero_K00 config flag)
        if conn.is_transmission:
            if self.sim.model_data.zero_K00:
                conn.bhelms_mtx_func = fast_compute_trns_knm_matrix_bh__k00_zeroed
            else:
                conn.bhelms_mtx_func = fast_compute_trns_knm_matrix_bh
        else:
            if self.sim.model_data.zero_K00:
                conn.bhelms_mtx_func = fast_compute_refl_knm_matrix_bh__k00_zeroed
            else:
                conn.bhelms_mtx_func = fast_compute_refl_knm_matrix_bh

    cdef initialise_knm_workspaces(KnmConnectorWorkspace self):
        cdef:
            NodeBeamParam *q_from
            NodeBeamParam *q_to
            KnmInfo *info
            # From Beam parameters transformed by ABCD
            complex_t qx_from_trns, qy_from_trns

            double lambda0 = self.sim.model_data.lambda0
            int maxtem = self.sim.model_data.maxtem
            int i = 0
            double beta_x, beta_y

        oconn_names = tuple(self.o2o.keys())
        # Loop over the node ids stored in input, output pairs and get the
        # beam param data
        for i in range(self.N_opt_conns):
            info = &self.oconn_info[i]
            if not info.has_been_set:
                raise Exception(f"Information for connection {oconn_names[i]} at {self.owner.name} has not been set with `set_knm_info`")

            if info.abcd_x == NULL:
                raise Exception(f"Knm info ABCDx for connection {oconn_names[i]} at {self.owner.name} is NULL")
            if info.abcd_y == NULL:
                raise Exception(f"Knm info ABCDy for connection {oconn_names[i]} at {self.owner.name} is NULL")
            if info.beta_x == NULL and info.beta_x_factor != 0:
                raise Exception(f"Knm info beta_x for connection {oconn_names[i]} at {self.owner.name} is NULL but has a beta_x_factor set")
            if info.beta_y == NULL and info.beta_y_factor != 0:
                raise Exception(f"Knm info beta_y for connection {oconn_names[i]} at {self.owner.name} is NULL but has a beta_y_factor set")

            ni_idx = self.onode_ids[2*info.from_port_idx]
            no_idx = self.onode_ids[2*info.to_port_idx+1]
            q_from = &self.sim.trace[ni_idx]
            q_to = &self.sim.trace[no_idx]
            # 'From' Beam parameters after propagation through the connections ABCD
            qx_from_trns = c_transform_q(info.abcd_x, q_from.qx.q, info.nr_from, info.nr_to)
            qy_from_trns = c_transform_q(info.abcd_y, q_from.qy.q, info.nr_from, info.nr_to)
            # Get any misalignment factor is one has been set
            beta_x = info.beta_x[0] if info.beta_x else 0
            beta_y = info.beta_y[0] if info.beta_y else 0
            # Initialise the Knm workspaces with the current values
            knm_bh_ws_init(
                info.K_ws_x, qx_from_trns, q_to.qx.q, beta_x, info.beta_x_factor, info.nr_to, lambda0, maxtem
            )
            knm_bh_ws_init(
                info.K_ws_y, qy_from_trns, q_to.qy.q, beta_y, info.beta_y_factor, info.nr_to, lambda0, maxtem,
            )

    cdef void free_knm_workspaces(KnmConnectorWorkspace self) nogil:
        cdef:
            int i
            knm_bh_workspace *kws = NULL

        for i in range(2*self.N_opt_conns):
            knm_bh_ws_free(&self.Kws[i])

    cdef void flag_changing_knm_workspaces(KnmConnectorWorkspace self):
        cdef:
            KnmInfo *info
            bint is_mm_changing
            Py_ssize_t ni_idx, no_idx
            const NodeBeamParam* q_from
            const NodeBeamParam* q_to

        for i in range(self.N_opt_conns):
            info = &self.oconn_info[i]
            ni_idx = self.onode_ids[2*info.from_port_idx]
            no_idx = self.onode_ids[2*info.to_port_idx+1]
            q_from = &self.sim.trace[ni_idx]
            q_to = &self.sim.trace[no_idx]

            # Mode mismatch is only changing if this node coupling is
            # in the pre-determined mismatched_node_couplings of the sim
            is_mm_changing = (
                list(self.o2o.values())[i] in self.sim.mismatched_node_couplings
                # Or if the component is misaligned and we have changing beam parameters
                # at this coupling then also need to recompute scatter matrix
                or (
                    (not info.K_ws_x.aligned or not info.K_ws_y.aligned)
                    and (q_from.is_changing or q_to.is_changing)
                )
            )
            info.K_ws_x.is_mm_changing = is_mm_changing
            info.K_ws_y.is_mm_changing = is_mm_changing
            info.K_ws_x.is_alignment_changing = info.beta_x_is_changing
            info.K_ws_y.is_alignment_changing = info.beta_y_is_changing

            if knm_bh_ws_is_changing(info.K_ws_x) or knm_bh_ws_is_changing(info.K_ws_y):
                LOGGER.debug(f"{self.owner.name}.K{info.from_port_idx+1}{info.to_port_idx+2} is changing", )

    cdef void update_changing_knm_workspaces(KnmConnectorWorkspace self) nogil:
        cdef:
            NodeBeamParam *q_from
            NodeBeamParam *q_to
            KnmInfo *info
            complex_t qx_from_trns, qy_from_trns

        for i in range(self.N_opt_conns):
            info = &self.oconn_info[i]
            ni_idx = self.onode_ids[2*info.from_port_idx]
            no_idx = self.onode_ids[2*info.to_port_idx+1]
            q_from = &self.sim.trace[ni_idx]
            q_to = &self.sim.trace[no_idx]

            # if there's some changing Knm then we need to recompute it. Which parts need
            # to be recomputed are decided here then performed
            if info.K_ws_x.is_mm_changing:
                qx_from_trns = c_transform_q(info.abcd_x, q_from.qx.q, info.nr_from, info.nr_to)
                if info.beta_x == NULL:
                    knm_bh_ws_recompute_mismatch(info.K_ws_x, qx_from_trns, q_to.qx.q)
                else:
                    knm_bh_ws_recompute(info.K_ws_x, qx_from_trns, q_to.qx.q, info.beta_x[0])

            elif info.K_ws_x.is_alignment_changing and info.beta_x != NULL:
                knm_bh_ws_recompute_misalignment(info.K_ws_x, info.beta_x[0])

            if info.K_ws_y.is_mm_changing:
                qy_from_trns = c_transform_q(info.abcd_y, q_from.qy.q, info.nr_from, info.nr_to)
                if info.beta_y == NULL:
                    knm_bh_ws_recompute_mismatch(info.K_ws_y, qy_from_trns, q_to.qy.q)
                else:
                    knm_bh_ws_recompute(info.K_ws_y, qy_from_trns, q_to.qy.q, info.beta_y[0])

            elif info.K_ws_y.is_alignment_changing and info.beta_y != NULL:
                knm_bh_ws_recompute_misalignment(info.K_ws_y, info.beta_y[0])

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void compute_scattering_matrices(KnmConnectorWorkspace self):
        cdef:
            KnmInfo *info
            Py_ssize_t i

        for i in range(self.N_opt_conns):
            info = &self.oconn_info[i]
            if knm_bh_ws_is_changing(info.K_ws_x) or knm_bh_ws_is_changing(info.K_ws_y):
                info.bhelms_mtx_func(
                    info.K_ws_x,
                    info.K_ws_y,
                    self.sim.model_data.homs_ptr,
                    info.mtx.ptr,
                    self.sim.model_data.num_HOMs,
                    self.sim.config_data.nthreads_homs,
                )

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void compute_knm_losses(KnmConnectorWorkspace self) nogil:
        cdef:
            KnmInfo *info
            Py_ssize_t i

        for i in range(self.N_opt_conns):
            info = &self.oconn_info[i]
            if knm_bh_ws_is_changing(info.K_ws_x) or knm_bh_ws_is_changing(info.K_ws_y):
                knm_loss(info.mtx.ptr, info.loss, self.sim.model_data.num_HOMs)


# N.B. This function depends on the order of ports / nodes being consistent between
# ws.onode_ids and ws.oconn_info, and may break if this order is changed.
# TODO: This doesn't consider the couplings in the signal matrix, effectively assuming they're 1,
# which isn't always correct.
optical_quantum_noise_knm = FillFuncWrapper.make_from_ptr(c_optical_quantum_noise_knm)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef object c_optical_quantum_noise_knm(ConnectorWorkspace cws):
    cdef:
        KnmConnectorWorkspace ws = <KnmConnectorWorkspace> cws
        NoiseInfo noises = ws.output_noise
        frequency_info_t *freq

        Py_ssize_t i, j, k, h
        Py_ssize_t port_idx

        complex_t factor

    if ws.sim.signal.nhoms == 0:
        return
    for i in range(ws.sim.signal.optical_frequencies.size):
        freq = &(ws.sim.signal.optical_frequencies.frequency_info[i])
        factor = 0.5 * (1 + freq.f_car[0] / ws.sim.model_data.f0)
        for j in range(noises.num_nodes):
            port_idx = -1
            for k in range(ws.N_opt_ports):
                if ws.onode_ids[2 * k + 1] == noises.node_info[j].idx:
                    port_idx = k
                    break
            if port_idx == -1:
                continue
            ws.total_losses[:] = 0
            for k in range(ws.N_opt_conns):
                if ws.oconn_info[k].to_port_idx == port_idx:
                    for h in range(ws.sim.signal.nhoms):
                        ws.total_losses[h] += ws.oconn_info[k].loss[h]
            (<SubCCSView>noises.ptrs[j][freq.index]).fill_za_dd(factor, ws.total_losses)



from finesse.cymath cimport complex_t
from finesse.cymath.complex cimport cexp, crotate
from finesse.cymath.math cimport sqrt, radians
from finesse.cymath.gaussbeam cimport bp_gouy
from finesse.cmatrix cimport SubCCSView, SubCCSView1DArray
from finesse.knm cimport KnmMatrix
from finesse.simulations.base cimport NodeBeamParam
from finesse.frequency cimport frequency_info_t
from finesse.cymath.math cimport sgn
from finesse.simulations.basematrix cimport MatrixSystemSolver, CarrierSignalMatrixSimulation

import numpy as np
cimport numpy as np
cimport cython

from cpython.ref cimport PyObject, Py_XINCREF, Py_XDECREF
from libc.string cimport strcmp, memcpy
from libc.stdlib cimport free, calloc

ctypedef (double*, double*, double*) ptr_tuple_3

cdef extern from "constants.h":
    long double PI
    double C_LIGHT
    double DEG2RAD

cdef class LaserConnections:
    def __cinit__(self, MatrixSystemSolver mtx):
        cdef:
            int Nfo =  mtx.optical_frequencies.size

        # There are no carrier connections at a laser, just signals
        self.SIGAMP_P1o = SubCCSView2DArray(1, Nfo)
        self.SIGFRQ_P1o = SubCCSView2DArray(1, Nfo)
        self.SIGPHS_P1o = SubCCSView2DArray(1, Nfo)

        self.ptrs.SIGAMP_P1o = <PyObject***>self.SIGAMP_P1o.views
        self.ptrs.SIGFRQ_P1o = <PyObject***>self.SIGFRQ_P1o.views
        self.ptrs.SIGPHS_P1o = <PyObject***>self.SIGPHS_P1o.views

cdef class LaserValues(BaseCValues):
    def __init__(self):
        cdef ptr_tuple_3 ptr = (&self.P, &self.phase, &self.f)
        cdef tuple params = ("P", "phase", "f")
        self.setup(params, sizeof(ptr), <double**>&ptr)


cdef class LaserWorkspace(ConnectorWorkspace):
    def __init__(self, object owner, CarrierSignalMatrixSimulation sim, bint refill):
        super().__init__(
                owner,
                sim,
                refill,
                refill,
                None,
                LaserConnections(sim.signal) if sim.signal else None,
                LaserValues()
                )
        self.v = self.values
        self.lc = self.signal.connections if sim.signal else None
        self.PIj_2 = PI*0.5j

        # indexes for beam tracing
        self.P1o_id = sim.carrier.node_id(owner.p1.o)


laser_carrier_fill_rhs = FillFuncWrapper.make_from_ptr(c_laser_carrier_fill_rhs)
cdef object c_laser_carrier_fill_rhs(ConnectorWorkspace cws):
    r"""
    Fills the right hand side (RHS) vector corresponding to the light source `laser`.

    The field amplitude is set as

    .. math::
        a_{\mathrm{in}} = \sqrt{\frac{2P}{\epsilon_c}}~\exp{\left(i \varphi\right)},

    where :math:`P` is the laser power and :math:`\varphi` is the specified phase of
    the laser.

    Parameters
    ----------

    laser : :class:`.Laser`
        The laser object to fill.

    sim : :class:`.CarrierSignalMatrixSimulation`
        A handle to the simulation.

    values : dict
        Dictionary of evaluated model parameters.

    fsrc_index : int
        Index of source frequency bin.
    """
    cdef:
        LaserWorkspace ws = <LaserWorkspace>cws

        # Carrier laser injection
        complex_t Ein = sqrt(2 * ws.v.P / ws.sim.model_data.EPSILON0_C) * cexp(1.0j * radians(ws.v.phase))

        Py_ssize_t i

    if not ws.sim.is_modal:
        ws.sim.carrier.set_source_fast(
            ws.node_car_id, ws.fsrc_car_idx, 0, Ein,
        )
    else:
        for i in range(ws.sim.model_data.num_HOMs):
            ws.sim.carrier.set_source_fast(
                ws.node_car_id, ws.fsrc_car_idx, i, Ein * ws.power_coeffs[i],
            )


laser_fill_qnoise = FillFuncWrapper.make_from_ptr(c_laser_fill_qnoise)
cdef object c_laser_fill_qnoise(ConnectorWorkspace cws):
    r"""
    Fills the quantum noise input matrix corresponding to the light source `laser`.
    """
    cdef:
        LaserWorkspace ws = <LaserWorkspace>cws
        PyObject ***noises = ws.output_noise.ptrs
        frequency_info_t *freq

        # Laser quantum noise injection
        complex_t qn

    for i in range(ws.sim.signal.optical_frequencies.size):
        freq = &(ws.sim.signal.optical_frequencies.frequency_info[i])
        qn = ws.sim.model_data.UNIT_VACUUM / 2 * (1 + freq.f_car[0] / ws.sim.model_data.f0)
        (<SubCCSView>noises[0][freq.index]).fill_za(qn)


laser_fill_signal = FillFuncWrapper.make_from_ptr(c_laser_fill_signal)
cdef object c_laser_fill_signal(ConnectorWorkspace cws):
    cdef:
        LaserWorkspace ws = <LaserWorkspace>cws
        laser_connections conns = <laser_connections>ws.lc.ptrs
        Py_ssize_t i, j
        double factor = ws.sim.model_data.EPSILON0_C

        complex_t phs_sig
        complex_t frq_sig

        frequency_info_t *f
        complex_t* car

    for i in range(2):
        f = &ws.sim.signal.optical_frequencies.frequency_info[ws.fcar_sig_sb_idx[i]]
        # NOTE shouldn't need multiplication by corresponding power_coeff
        #      factor as this was taken into account in DC computations
        # Update this so that frequency rhs idx is store in workspace
        j = ws.sim.carrier.field_fast(ws.node_car_id, f.audio_carrier_index, 0)
        car = <complex_t*>&(ws.sim.carrier.out_view[j])
        # TODO ddb - these are all assuming a single electronic frequency here
        if conns.SIGAMP_P1o[0][f.index]:
            (<SubCCSView>conns.SIGAMP_P1o[0][f.index]).fill_negative_za_zm_2(factor * 0.5 * 0.5, car, 1, 1)

        if conns.SIGFRQ_P1o[0][f.index]:
            frq_sig =  0.5 / ws.sim.model_data.fsig * sgn(f.audio_order)
            (<SubCCSView>conns.SIGFRQ_P1o[0][f.index]).fill_negative_za_zm_2(factor * frq_sig, car, 1, 1)

        if conns.SIGPHS_P1o[0][f.index]:
            phs_sig = 1j * 0.5
            (<SubCCSView>conns.SIGPHS_P1o[0][f.index]).fill_negative_za_zm_2(factor * phs_sig, car, 1, 1)


laser_set_gouy = GouyFuncWrapper.make_from_ptr(set_tem_gouy_phases)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.initializedcheck(False)
cdef int set_tem_gouy_phases(ConnectorWorkspace ws) except -1:
    cdef:
        const NodeBeamParam* q_p1o = &ws.sim.trace[ws.P1o_id]

        double gouy_x
        double gouy_y
        double phase00 = 0.0
        double phase

        Py_ssize_t i
        int n, m

    if not q_p1o.is_changing:
        return 0

    gouy_x = bp_gouy(&q_p1o.qx)
    gouy_y = bp_gouy(&q_p1o.qy)

    if ws.sim.model_data.zero_tem00_gouy:
        phase00 = 0.5 * gouy_x + 0.5 * gouy_y

    for i in range(ws.sim.model_data.num_HOMs):
        n = ws.sim.model_data.homs_view[i][0]
        m = ws.sim.model_data.homs_view[i][1]

        phase = (n + 0.5) * gouy_x + (m + 0.5) * gouy_y
        ws.power_coeffs[i] = crotate(ws.power_coeffs[i], phase - phase00)

    return 0



from libc.stdlib cimport free, calloc
from libc.string cimport memcpy
from cpython.ref cimport Py_XINCREF, Py_XDECREF

from collections import defaultdict, ChainMap
import weakref
from copy import deepcopy
import logging

from .parameter cimport (
    Parameter,
    ParameterState,
    GeometricParameter,
)
from .exceptions import ModelParameterDefaultValueError


LOGGER = logging.getLogger(__name__)


class ModelElement:
    """Base for any object which can be an element of a :class:`.Model`.

    When added to a model it will attempt to call the method `_on_add` so that the element can do
    some initialisation if required.

    Parameters
    ----------
    name : str
        Name of newly created model element.
    """

    # A global dictionary to keep a record of all the declared
    # model parameters, validators, etc.
    _param_dict = defaultdict(list)
    _validators = defaultdict(dict)
    _setters = defaultdict(dict)
    _default_parameter_name = dict()

    # Info parameters.
    _info_param_dict = defaultdict(dict)

    def __new__(cls, name, *args, **kwargs):
        instance = super(ModelElement, cls).__new__(cls)
        instance._params = []
        instance._info_params = instance._info_param_dict[cls]
        instance._ModelElement__name = name

        # Loop through each of the parameters that have been defined
        # in the class and instantiate an object to represent them
        # for this instance of the object
        for pinfo in instance._param_dict[cls]:
            if pinfo.is_geometric:
                p = GeometricParameter(pinfo, instance)
            else:
                p = Parameter(pinfo, instance)
            setattr(instance, f"__param_{pinfo.name}", p)
            instance._params.append(p)

        return instance

    def __init__(self, name):
        from finesse.utilities import check_name

        self.__model = None
        self.__name = check_name(name)
        self._add_to_model_namespace = True
        self._namespace = "."

    def __str__(self):
        params = {param.name: str(param.value) for param in self.parameters}
        info_params = {param: str(getattr(self, param)) for param in self._info_params}
        values = [repr(self.name)] + [f"{k}={v}" for k, v in ChainMap(info_params, params).items()]
        return f"{self.__class__.__name__}({', '.join(values)})"

    def __repr__(self):
        return "<'{}' @ {} ({})>".format(
            self.name, hex(id(self)), self.__class__.__name__
        )

    def __deepcopy__(self, memo):
        new = object.__new__(type(self))
        memo[id(self)] = new

        # For debugging what causes deepcopy errors
        # try:
        #     for key in self.__dict__:
        #         new.__dict__[key] = deepcopy(self.__dict__[key], memo)
        # except Exception:
        #     print("ERROR on deepcopy", key)
        #     raise

        new.__dict__.update(deepcopy(self.__dict__, memo))

        # Manually update the weakrefs to be correct
        new.__model = weakref.ref(memo[id(self.__model())])
        return new

    def info(self, tablefmt="fancy_grid"):
        """Element information.

        Parameters
        ----------
        tablefmt : str, optional; default: "fancy_grid"
            The plain-text table format to pass to tabulate.

        Returns
        -------
        str
            The formatted info.
        """
        params = self.parameter_table(tablefmt=tablefmt)
        info_params = self.info_parameter_table(tablefmt=tablefmt)

        msg = f"{self.__class__.__name__} {self.name}\n"
        msg += "\nParameters:\n"
        if params is not None:
            msg += params
        else:
            msg += "n/a\n"

        if info_params is not None:
            msg += "\nInformation:\n"
            msg += self.info_parameter_table(tablefmt=tablefmt)

        return msg

    def parameter_table(self, tablefmt="fancy_grid"):
        """Model parameter table.

        Parameters
        ----------
        tablefmt : str, optional; default: "fancy_grid"
            The plain-text table format to pass to tabulate.

        Returns
        -------
        str
            The formatted parameter info table.
        None
            If there are no parameters.
        """
        from tabulate import tabulate

        if not self.parameters:
            return

        table = []
        headers = ["Description", "Value"]

        # Loop in reverse so we can keep the natural order of each element's parameters in its
        # corresponding model parameter class decorators.
        for field in reversed(self.parameters):
            table.append([field.description, field])

        return tabulate(table, headers, tablefmt=tablefmt)

    def info_parameter_table(self, tablefmt="fancy_grid"):
        """Info parameter table.

        This provides a table with useful fields in addition to those contained in
        :meth:`.parameter_table`.

        Parameters
        ----------
        tablefmt : str, optional; default: "fancy_grid"
            The plain-text table format to pass to tabulate.

        Returns
        -------
        str
            The formatted extra info table.
        None
            If there are no info parameters.
        """
        from tabulate import tabulate

        if not self._info_params:
            return

        # Loop in reverse so we can keep the natural order of each element's parameters in its
        # corresponding info parameter class decorators.
        table = [
            [description, getattr(self, name)]
            for name, (description, _) in reversed(self._info_params.items())
        ]
        headers = ["Description", "Value"]

        return tabulate(table, headers, tablefmt=tablefmt)

    @property
    def parameters(self):
        """Returns a list of the parameters available for this element"""
        return self._params.copy()

    @property
    def default_parameter_name(self):
        """The default parameter to assume when the component is directly referenced.

        This is used for example in kat script when the component is directly referenced in an
        expression, instead of the model parameter, e.g. &l1 instead of &l1.P.

        Returns
        -------
        str
            The name of the default model parameter.

        None
            If there is no default.
        """
        return self._default_parameter_name.get(self.__class__)

    @property
    def name(self):
        """Name of the element.

        Returns
        -------
        str
            The name of the element.
        """
        return self.__name

    @property
    def ref(self):
        """Reference to the default model parameter, if set.

        Returns
        -------
        :class:`.ParameterRef`
            Reference to the default model parameter, if set.

        Raises
        ------
        ValueError
            If there is no default model parameter set for this element.
        """
        if self.default_parameter_name is None:
            raise ModelParameterDefaultValueError(self)

        return getattr(self, self.default_parameter_name).ref

    @property
    def _model(self):
        """
        Internal reference to the model this element has been added to.

        Raises
        ------
        ComponentNotConnected when not connected
        """
        from finesse.exceptions import ComponentNotConnected

        if self.__model is None:
            raise ComponentNotConnected(f"{self.name} is not connected to a model")
        else:
            return self.__model()

    @property
    def has_model(self):
        """Returns true if this element has been associated with a Model."""
        return self.__model is not None

    def _set_model(self, model):
        """
        A :class:`.Model` instance calls this to associate itself
        with the element.

        .. note::
            This method should never be called by the user, it should
            only be called internally by the :class:`.Model` class.

        Parameters
        ----------
        model : :class:`.Model`
            The model to associate with this element.

        Raises
        ------
        Exception
            If the model is already set for this element.
        """
        if self.__model is not None:
            raise Exception("Model is already set for this element")

        self.__model = weakref.ref(model)

    def _reset_model(self, new_model):
        """Resets the model that this element is associated with. Note, this should
        not be used in normal coding situations. It should only be used when writing
        new elements that override the `__deepcopy__` method.
        """
        self.__model = weakref.ref(new_model)

    def _setup_changing_params(self):
        """For any parameter that has been set to be changing during a simulation
        this method will store them and their evaluated values in the set `self._params_changing`
        and the dict `self._params_evald`.
        """
        # N.B. Decorators are evaluated from inside - out, so to make the order returned here match
        # the order defined, we must reverse self.parameters
        self._params_changing = set(
            p for p in reversed(self.parameters) if p.is_changing
        )
        self._params_evald = {
            p.name: (p.value.eval() if hasattr(p.value, "eval") else p.value)
            for p in reversed(self.parameters)
        }

    def _clear_changing_params(self):
        """Sets the set `self._params_changing` and the dict `self._params_evald` to None
        after a simulation has completed."""
        self._params_changing = None
        self._params_evald = None

    def _eval_parameters(self):
        """
        Only call this methods when the model is built. It is optimised
        for returned changed parameter values in this state.

        To get a dictionary of parameter values in other cases use:

        >>>> values = {p.name:p.eval() for p in element.parameters}

        Returns
        -------
        params : dict(str:float)
            Dictionary of parameter values
        params_changing : set(str)
            A set of parameter names which are changing.
        """
        # Now we know which are evaluable so no need for repeated checks
        # Also reduce memory bashing creating a ton of dictionaries and
        # reuse just one.
        for p in self._params_changing:
            self._params_evald[p.name] = p.eval()

        return self._params_evald, self._params_changing


class ElementValues:
    """Standard Python object which is used to store an Elements Parameter values.
    This is used in the default case where no optimised C class is provided."""
    pass


cdef class BaseCValues:
    """Base class that elements should use for storing their parameter
    values. This contains a generic method for storing a pointers to the
    double parameter values, which is intialised using `setup()`.

    Each Parameter of an element should result in a double. This won't handle
    any other types of data.
    """
    def __cinit__(self):
        self.ptr = NULL
        self.N = 0

    cdef setup(self, tuple params, Py_ssize_t data_size, double** data_start):
        """Allocates the memory needed to store double values for the requested parameters.

        Example
        -------
        An array of pointers to the double variables where each parameter should be copied across
        to.

        ctypedef (double*, double*) ptr_tuple_2 # data type of double points

        cdef class OpticalBandpassValues(BaseCValues):
            def __init__(self):
                cdef ptr_tuple_2 ptr = (&self.fc, &self.bandwidth) # array of ptrs to store parameter values at
                cdef tuple params = ("fc", "bandwidth") # names that match up in order with pointers
                self.setup(params, sizeof(ptr), <double**>&ptr) # call setup

        Parameters
        ----------
        params : tuple
            Tuple of Parameter names (case-sensitive) that an element has and needs to be stored in C values
        data_size : unsigned long
            number of double* pointers (or number of parameters)
        data_start : double**
            Pointer to array of double pointers where each parameter value should be stored
        """
        if self.ptr != NULL:
            raise ValueError("Memory already allocated")
        if data_start == NULL:
            raise ValueError("data_start == NULL")

        self.params = params
        self.N = data_size//sizeof(double)

        if len(self.params) != self.N:
            raise ValueError("Tuple of parameters and length of double pointers does not match")

        self.ptr = <double**> calloc(self.N, sizeof(double*))

        if not self.ptr:
            raise MemoryError()

        memcpy(self.ptr, data_start, data_size)

    def __dealloc__(self):
        if self.ptr != NULL:
            free(self.ptr)
            self.ptr = NULL

    cdef get_param_ptr(self, unicode name, double**ptr):
        """Get a C pointer to where an elements parameter values should be stored.

        Parameters
        ----------
        name : unicode
            Name of the parameter
        ptr : double*
            Pointer to where the parameter pointer should be stored
        """
        if self.ptr == NULL:
            raise ValueError("Value storage pointers not set")
        if ptr == NULL:
            raise ValueError("ptr is NULL")

        idx = self.params.index(name)
        if self.ptr[idx] == NULL:
            raise ValueError(f"Pointer to {name} double is NULL")

        ptr[0] = self.ptr[idx]


cdef class ElementWorkspace:
    """
    This is the most basic workspace for a model element. It keeps track of the owner element and
    its parameter values in the `self.values` object.
    """
    def __cinit__(self, *args, **kwargs):
        self.num_changing_parameters = 0
        self.chprm = NULL
        self.chprm_target = NULL
        self.chprm_expr = NULL
        self.owner_id = -1
        self.first = True

    def __init__(self, object owner, object values=None):
        cdef:
            Parameter p
            int i
            object lmd
        self.owner = owner

        if values is None:
            self.values = ElementValues()
            self.type_c_values = None
            self.is_c_values = False
        else:
            if not isinstance(values, BaseCValues):
                raise Exception("Values object should be a derivative of BaseCValues")

            self.values = values
            self.type_c_values = type(values)
            self.is_c_values = True

            # Here we setup for fast parameter setting by storing
            # the pyobject pointers to Parameters and also a double
            # pointer to the value

            # Make sure that numeric parameters come before symbolic
            # parameters so that the latter get eval'd to the correct
            # value when calling cy_expr_eval
            numeric_params = []
            symbolic_params = []
            for p in owner.parameters:
                if p.is_changing:
                    self.num_changing_parameters += 1
                    if p.state == ParameterState.Symbolic:
                        symbolic_params.append(p)
                    else:
                        numeric_params.append(p)

            params = numeric_params + symbolic_params
            if self.num_changing_parameters > 0:
                self.chprm = <PyObject**> calloc(self.num_changing_parameters, sizeof(PyObject*))
                self.chprm_target = <double**> calloc(self.num_changing_parameters, sizeof(double*))
                self.chprm_expr = <cy_expr**> calloc(self.num_changing_parameters, sizeof(cy_expr*))
                i = 0
                for p in params:
                    if p.is_changing:
                        if p.state == ParameterState.Symbolic:
                            self.chprm_expr[i] = cy_expr_new()

                        self.chprm[i] = <PyObject*>p
                        Py_XINCREF(self.chprm[i])
                        (<BaseCValues>self.values).get_param_ptr(p.name, &self.chprm_target[i])
                        i += 1

    cpdef int compile_cy_exprs(self) except -1:
        if not self.num_changing_parameters:
            return 0

        for i in range(self.num_changing_parameters):
            if (<Parameter>self.chprm[i]).state == ParameterState.Symbolic:
                if cy_expr_init(
                    self.chprm_expr[i],
                    (<Parameter>self.chprm[i]).value.expand_symbols().eval(keep_changing_symbols=True)
                ) == -1:
                    return -1

        return 0

    cpdef update_parameter_values(self):
        """Updates the `self.values` container which holds the latest values
        of this elements parameters.
        """
        cdef unicode p

        if self.first or not self.is_c_values:
            # First go just fill everything.
            # This is fairly slow but robust for any python object...
            vals, _ = self.owner._eval_parameters()
            for p in vals:
                if vals[p] is not None:
                    setattr(self.values, p, vals[p])
                else:
                    setattr(self.values, p, 0)

            self.first = False

        elif self.num_changing_parameters > 0:
            # Here we do some lower level setting of parameters using pointers
            # to the workspace.value.parameter double variable to speed things up
            for i in range(self.num_changing_parameters):
                if self.chprm_target[i] == NULL or self.chprm[i] == NULL:
                    raise MemoryError()

                if (<Parameter>self.chprm[i]).state == ParameterState.Numeric:
                    self.chprm_target[i][0] = (<Parameter>self.chprm[i]).__cvalue
                elif (<Parameter>self.chprm[i]).state == ParameterState.NONE:
                    self.chprm_target[i][0] = 0 # maybe should be NaN?
                elif (<Parameter>self.chprm[i]).state == ParameterState.Symbolic:
                    self.chprm_target[i][0] = cy_expr_eval(self.chprm_expr[i])
                    # Make sure __cvalue of symbolic parameters get updated too so that
                    # anything that relies on address of this uses correct value
                    (<Parameter>self.chprm[i]).__cvalue = self.chprm_target[i][0]
                else:
                    raise ValueError("Parameter state was unexpected")

    def __dealloc__(self):
        if self.num_changing_parameters > 0:
            for i in range(self.num_changing_parameters):
                if (<Parameter>self.chprm[i]).state == ParameterState.Symbolic:
                    cy_expr_free(self.chprm_expr[i])

                if self.chprm[i] == NULL:
                    raise MemoryError()
                Py_XDECREF(self.chprm[i])

            if self.chprm != NULL:
                free(self.chprm)
                self.chprm = NULL
            if self.chprm_expr != NULL:
                free(self.chprm_expr)
                self.chprm_expr = NULL
            if self.chprm_target != NULL:
                free(self.chprm_target)
                self.chprm_target = NULL



cimport numpy as np
from finesse.cymath.complex cimport complex_t, DenseZMatrix


cdef class KnmMatrix:
    cdef readonly:
        np.ndarray data
        str name

    cdef:
        complex_t[:, ::1] data_view
        DenseZMatrix mtx
        const int[:, ::1] modes_view

    cdef (Py_ssize_t, Py_ssize_t) field_indices_from(self, key)
    cdef complex_t coupling(self, int n1, int m1, int n2, int m2) nogil


cpdef make_unscaled_X_scatter_knm_matrix(int[:,::1] modes)
cpdef make_unscaled_Y_scatter_knm_matrix(int[:,::1] modes)


# Compute loss from scattering for each coupling,
# required for  quantum noise calculations
cdef void knm_loss(const complex_t* knm_mat, double* out, Py_ssize_t N) nogil



cimport numpy as np
from cpython.ref cimport PyObject
from finesse.cymath cimport complex_t
from cython cimport view

from finesse.cymath.complex cimport DenseZVector, DenseZMatrix

ctypedef np.npy_intp SIZE_t

cdef extern from "klu.h":
    cdef int KLU_OK = 0
    cdef int KLU_SINGULAR = 1
    cdef int KLU_OUT_OF_MEMORY = -2
    cdef int KLU_INVALID = -3
    cdef int KLU_TOO_LARGE = -4

    ctypedef long SuiteSparse_long

    ctypedef struct klu_l_symbolic:
        pass

    ctypedef struct klu_l_numeric:
        pass

    ctypedef struct klu_l_common:
        int ordering
        int scale
        int btf
        int maxwork
        int status
        double rcond
        double condest
        double rgrowth

    cdef int klu_l_defaults(klu_l_common* Common)
    cdef klu_l_symbolic *klu_l_analyze(SuiteSparse_long num_eqs,
                                       SuiteSparse_long *col_ptr,
                                       SuiteSparse_long *row_idx,
                                       klu_l_common* Common)

    cdef klu_l_numeric *klu_zl_factor(SuiteSparse_long *col_ptr,
                                      SuiteSparse_long *row_idx,
                                      double *values,
                                      klu_l_symbolic *Symbolic,
                                      klu_l_common *Common)


    cdef void        klu_zl_refactor(SuiteSparse_long *col_ptr,
                                     SuiteSparse_long *row_idx,
                                     double *values,
                                     klu_l_symbolic *Symbolic,
                                     klu_l_numeric *Numeric,
                                     klu_l_common *Common)


    cdef int klu_zl_sort(klu_l_symbolic *Symbolic, klu_l_numeric *Numeric, klu_l_common* Common)

    cdef int klu_zl_free_numeric(klu_l_numeric **Numeric, klu_l_common *Common)
    cdef int klu_l_free_symbolic(klu_l_symbolic **Symbolic, klu_l_common *Common)

    cdef SuiteSparse_long klu_zl_solve (klu_l_symbolic *, klu_l_numeric *, SuiteSparse_long, SuiteSparse_long, double *, klu_l_common *)
    cdef SuiteSparse_long klu_zl_tsolve(klu_l_symbolic *, klu_l_numeric *, SuiteSparse_long, SuiteSparse_long, double *, SuiteSparse_long, klu_l_common * )

    SuiteSparse_long klu_zl_rgrowth (SuiteSparse_long *, SuiteSparse_long *, double *, klu_l_symbolic *, klu_l_numeric *, klu_l_common *)
    SuiteSparse_long klu_zl_condest (SuiteSparse_long *, double *, klu_l_symbolic *, klu_l_numeric *, klu_l_common *)
    SuiteSparse_long klu_zl_rcond (klu_l_symbolic *, klu_l_numeric *, klu_l_common *)

cdef class CCSMatrix:
    cdef:
        unicode __name
        dict __indexes

        SuiteSparse_long num_eqs
        int allocated
        int num_nodes
        unsigned num_rhs

        SuiteSparse_long   *row_idx
        SuiteSparse_long   *col_ptr
        complex_t *values
        complex_t *rhs
        complex_t[:, ::1] rhs_view

        SuiteSparse_long   __nnz
        readonly dict sub_columns
        dict diag_map
        list __callbacks
        object __weakref__

    cpdef add_diagonal_elements(
        self,
        SuiteSparse_long Neqs,
        SuiteSparse_long index,
        unicode name,
        is_diagonal = ?
    )
    cpdef add_submatrix(self, SuiteSparse_long _from, SuiteSparse_long _to,
                        unicode name, callback=?, type_=?)
    cpdef void set_rhs(self, SuiteSparse_long index, complex_t value, unsigned rhs_index=?)
    cdef unsigned request_rhs_view(self)
    cdef complex_t[::1] get_rhs_view(self, unsigned index)
    cpdef construct(self, complex_t diagonal_fill=?)
    cdef np.ndarray get_matrix(self, SuiteSparse_long _from, SuiteSparse_long _to, complex_t** start_ptr, SuiteSparse_long* from_rhs_index)
    cpdef clear_rhs(self, unsigned rhs_index=?)
    cpdef factor(self)
    cpdef refactor(self)
    cpdef const complex_t[::1] solve(self, int transpose=?, bint conjugate=?, unsigned rhs_index=?)
    cpdef void solve_extra_rhs(self, int transpose=?, bint conjugate=?)
    cdef void zgemv(self, complex_t[::1] out, unsigned rhs_index=?)


cdef class SubCCSView:
    cdef:
        readonly object M
        readonly Py_ssize_t _from, _to
        readonly bint conjugate_fill
        readonly int stride1 # in units of 16 bytes
        readonly int stride2 # in units of 16 bytes
        readonly int size1
        readonly int size2
        readonly SuiteSparse_long start_idx
        readonly SuiteSparse_long from_rhs_index
        complex_t* ptr
        np.ndarray A
        readonly complex_t[:, ::1] from_rhs_view # rhs[rhs index, rhs values]
        tuple __shape
        tuple __strides

    cdef void fill_za(self, complex_t a)
    cdef void fill_zd(self, complex_t[::1] D)
    cdef void fill_dd(self, double[::1] D)
    cdef void fill_za_dd(self, complex_t a, double[::1] D)
    cdef void fill_zd_2(self, const complex_t* D, int s1) nogil
    cdef void fill_za_zd_2(self, complex_t a, const complex_t* D, int stride) nogil
    cdef void fill_za_zm(self, complex_t a, complex_t[:,::1] M)
    cdef void fill_za_zm_2(self, complex_t a, const complex_t* M, int s1, int s2)
    cdef void fill_za_zmc(self, complex_t a, const complex_t* M, int s1, int s2)


    cdef void fill_zm(self, complex_t[:, ::1] M)
    cdef void fill_negative_za(self, complex_t a)
    cdef void fill_negative_zd(self, complex_t[::1] D)
    cdef void fill_negative_dd(self, double[::1] D)
    cdef void fill_negative_za_dd(self, complex_t a, double[::1] D)
    cdef void fill_negative_zd_2(self, const complex_t* D, int s1) nogil
    cdef void fill_negative_za_zd_2(self, complex_t a, const complex_t* D, int stride) nogil
    cdef void fill_negative_za_zm(self, complex_t a, complex_t[:,::1] M)
    cdef void fill_negative_za_zm_2(self, complex_t a, const complex_t* M, int s1, int s2)
    cdef void fill_negative_za_zmc(self, complex_t a, const complex_t* M, int s1, int s2)
    cdef void fill_negative_zm(self, complex_t[:, ::1] M)

    cdef void fill_za_zmv(self, complex_t a, DenseZMatrix* M, DenseZVector* V)
    cdef void fill_negative_za_zmv(self, complex_t a, DenseZMatrix* M, DenseZVector* V)
    cdef void fill_za_zmvc(self, complex_t a, DenseZMatrix* M, DenseZVector* V)
    cdef void fill_negative_za_zmvc(self, complex_t a, DenseZMatrix* M, DenseZVector* V)

    cdef void fill_prop_za_zm(self, SubCCSView V, Py_ssize_t rhs_idx, complex_t a, complex_t[:,::1] M, bint increment)
    cdef void fill_prop_za(self, SubCCSView V, Py_ssize_t rhs_idx, complex_t a, bint increment)
    cdef void fill_neg_prop_za_zm(self, SubCCSView V, Py_ssize_t rhs_idx, complex_t a, complex_t[:,::1] M, bint increment)
    cdef void fill_neg_prop_za(self, SubCCSView V, Py_ssize_t rhs_idx, complex_t a, bint increment)


cdef class SubCCSView1DArray:
    cdef PyObject** views
    cdef readonly Py_ssize_t size


cdef class SubCCSView2DArray:
    cdef PyObject*** views
    cdef readonly (Py_ssize_t, Py_ssize_t) shape
    cdef readonly Py_ssize_t rows
    cdef readonly Py_ssize_t cols


cdef class KLUMatrix(CCSMatrix):
    cdef:
        klu_l_common Common
        klu_l_numeric* Numeric
        klu_l_symbolic* Symbolic

    cpdef factor(self)
    cpdef refactor(self)
    cpdef const complex_t[::1] solve(self, int transpose=?, bint conjugate=?, unsigned rhs_index=?)
    cpdef void solve_extra_rhs(self, int transpose=?, bint conjugate=?)
    cpdef double rgrowth(self)
    cpdef double rcond(self)
    cpdef double condest(self)
    cpdef void zgemv(self, complex_t[::1] out, unsigned rhs_index=?)



# cython profile=True
"""
Sparse matrix objects with factorisation and solving routines performed via KLU.
"""
import weakref

import numpy as np
cimport numpy as np
import cython
from cython cimport view
from cpython.ref cimport PyObject, Py_XINCREF, Py_XDECREF
from libc.stdlib cimport calloc, free, malloc, realloc
from libc.string cimport memset
from cython.operator cimport dereference as deref
from libc.stdio cimport printf
from finesse.cymath.complex cimport conj
from finesse.cymath cimport complex_t # type: np.complex128_t (i.e. double complex)

cdef sortkey(x):
    return int(x.to)

cdef status_string(int status):
    if status == KLU_SINGULAR:
        return "KLU_SINGULAR"
    elif status == KLU_OUT_OF_MEMORY:
        return "KLU_OUT_OF_MEMORY"
    elif status == KLU_INVALID:
        return "KLU_INVALID"
    elif status == KLU_TOO_LARGE:
        return "KLU_TOO_LARGE"
    else:
        return "UNKNOWN"

cdef class _Column:
    cdef public:
        Py_ssize_t start
        Py_ssize_t size
        Py_ssize_t index
        unicode name
        list submatrices

    def __init__(self, size, index, name):
        assert(size>0)
        self.size = size
        self.index = index
        self.name = name
        self.submatrices = []


cdef class _SubMatrix:
    cdef public:
        Py_ssize_t to
        Py_ssize_t rows
        Py_ssize_t columns
        unicode name
        unicode type

    def __init__(self, type_, from_size, to_size, to, name):
        assert(from_size>0)
        assert(to_size>0)
        self.to = to
        self.type = type_
        self.name = name
        self.columns = from_size #
        self.rows = to_size


cdef class CCSMatrix:
    def __init__(self, name):
        self.__name = name
        self.__indexes = {}
        self.sub_columns = {}
        self.num_nodes   = 0
        self.num_eqs     = 0
        self.num_rhs     = 1
        self.allocated   = 0

        self.values  = NULL
        self.col_ptr = NULL
        self.row_idx = NULL
        self.diag_map = {}
        self.__nnz  = 0
        self.__callbacks = []

    @property
    def num_equations(self): return int(self.num_eqs)

    @property
    def num_rhs(self): return int(self.num_rhs)

    @property
    def indexes(self): return self.__indexes

    @property
    def name(self): return self.__name

    cdef unsigned request_rhs_view(self):
        if self.rhs:
            raise Exception("Can't request rhs_view after the matrix has been built")
        self.num_rhs += 1
        return self.num_rhs - 1

    def get_sub_matrix_view(self, Py_ssize_t from_node, Py_ssize_t to_node, unicode name, bint conjugate_fill):
        mat = SubCCSMatrixView(self, from_node, to_node, name, conjugate_fill)
        self.add_submatrix(from_node, to_node, name, mat, type_='m')
        return mat

    def get_sub_diagonal_view(self, Py_ssize_t from_node, Py_ssize_t to_node, unicode name, bint conjugate_fill):
        mat = SubCCSMatrixViewDiagonal(self, from_node, to_node, name, conjugate_fill)
        self.add_submatrix(from_node, to_node, name, mat, type_='d')
        return mat

    def __dealloc__(self):
        if self.col_ptr: free(self.col_ptr)
        if self.row_idx: free(self.row_idx)
        if self.values:  free(self.values)
        if self.rhs:     free(self.rhs)

    cpdef add_diagonal_elements(
        self,
        SuiteSparse_long Neqs,
        SuiteSparse_long index,
        unicode name,
        is_diagonal = True
    ):
        """
        Adds a submatrix to the matrix along its diagonal. This defines what equations exist
        in the matrix, the submatrix values default to `1`, but a diagonal view is returned that
        allows them to be modified. Before other submatrices can be added to the matrix the
        diagonal must be specfied and how many equations it represents.

        Parameters
        ----------
        Neqs : Py_ssize_t
            Number of equations this submatrix represents
        _index : long
            Subcolumn index
        name : unicode
            Name used to indentify this coupling in the matrix for debugging
        """
        if index in self.sub_columns:
            raise Exception("Diagonal elements already specified at index {}".format(index))

        self.sub_columns[index] = _Column(Neqs, index, name)
        if is_diagonal:
            self.sub_columns[index].submatrices.append(_SubMatrix('d', Neqs, Neqs, index, name))
            mat = SubCCSMatrixViewDiagonal(self, index, index, name, False)
        else:
            self.sub_columns[index].submatrices.append(_SubMatrix('m', Neqs, Neqs, index, name))
            mat = SubCCSMatrixView(self, index, index, name, False)
        # Record what RHS index/number of equations this submatrix will start in
        self.diag_map[index] = self.num_eqs
        self.num_eqs += Neqs

        self.__callbacks.append(mat)
        return mat

    cpdef add_submatrix(self, SuiteSparse_long _from, SuiteSparse_long _to,
                        unicode name, callback=None, type_='m'):
        """
        Adds a submatrix to the matrix. The nomenclature of `_from` and `_to` refer to the variable
        dependency of the equations this submatrix represents, i.e. the equations in submatrix
        `_to` depends on the values in `-from`. Therefore `_from` is the subcolumn index and `_to`
        is the subrow index.

        Parameters
        ----------
        _from : long
            Subcolumn index
        _to : long
            Subcolumn index
        name : unicode
            Name used to indentify this coupling in the matrix for debugging
        callback : function()
            A callback function that will be called once the matrix has been constructed
        type_ : char, optional
            Either 'm' for a full submatrix or 'd' for a diagonal element only submatrix
        """
        if _from not in self.sub_columns:
            raise Exception("Must add a diagonal submatrix at index {} first for this subcolumn".format(_from))
        if _to   not in self.sub_columns:
            raise Exception("Must add a diagonal submatrix at index {} first for this subcolumn".format(_to))

        _to_size = self.sub_columns[_to].size
        _from_size = self.sub_columns[_from].size

        self.sub_columns[_from].submatrices.append(_SubMatrix(type_, _from_size, _to_size, _to, name))

        if callback: self.__callbacks.append(callback)

    cpdef void set_rhs(self, SuiteSparse_long index, complex_t value, unsigned rhs_index=0):
        """
        Sets the value of the entry at position `index` of the `rhs_index`th right-hand-side vector
        to `value`.

        Parameters
        ----------
        index : long
            The index in the rhs vector to set
        value : complex_t
            The value to set
        rhs_index : unsigned, optional
            Which rhs vector to change; defaults to 0
        """
        if rhs_index >= self.num_rhs:
            raise ValueError(f"Invalid rhs index {rhs_index}")
        self.rhs[rhs_index * self.num_eqs + index] = value

    cpdef construct(self, complex_t diagonal_fill=complex(1, 0)):
        """
        Constructing the matrix involves taking the metadata submatrix positions
        throughout the matrix and allocating the memory and building the various
        CCS matrix structures. After this the matrix can be populated and solved.

        Parameters
        ----------
        diagonal_fill : complex_t, optional
            Value to fill the diagonal of the matrix with; defaults to 1+0j
        """

        cdef:
            SuiteSparse_long i      = 0
            SuiteSparse_long j      = 0
            SuiteSparse_long k      = 0
            SuiteSparse_long cnnz   = 0 # current element number
            SuiteSparse_long crow   = 0 # current row
            SuiteSparse_long ccol   = 0 # current column in a submatrix
            SuiteSparse_long nsc    = len(self.sub_columns) # number of sub_columns
            SuiteSparse_long nnz    = 0
            _Column col
            _SubMatrix sm

        for col in self.sub_columns.values():
            # sort so that the the submatrices are in row order
            # ddb: This sort takes up about half the time of this method
            #      maybe some optimised instertion sorted list might be better
            col.submatrices.sort(key=sortkey)
            assert(col.size > 0)
            for sm in col.submatrices:
                # count how many elements per column for allocating memory
                if sm.type == 'm': # matrix
                    nnz += sm.rows * sm.columns
                elif sm.type == 'd': # diagonal
                    nnz += col.size
                else:
                    raise Exception("Unhandled")

        assert(self.num_eqs > 0)

        self.col_ptr  = <SuiteSparse_long*> malloc(sizeof(SuiteSparse_long) * (self.num_eqs + 1))
        self.row_idx  = <SuiteSparse_long*> malloc(sizeof(SuiteSparse_long) * nnz)
        self.values   = <complex_t*>        calloc(nnz, sizeof(complex_t))
        self.rhs      = <complex_t*>       calloc(self.num_eqs * self.num_rhs, sizeof(complex_t))
        self.rhs_view = <complex_t[:self.num_rhs, :self.num_eqs]>self.rhs

        for i, col in enumerate(self.sub_columns.values()): # For each subcolumn...
            col.start = cnnz # record index where this column starts

            for j in range(col.size): # then for each actual column in the subcolumn
                # set the starting location of the column in the pointer vector
                self.col_ptr[ccol] = cnnz

                for sm in col.submatrices: # select each submatrix...
                    crow = self.diag_map[sm.to]

                    # then set the elements in the column for...
                    if sm.type == 'm': # a matrix
                        for k in range(sm.rows):
                            self.row_idx[cnnz+k] = crow + k
                            # Set a default position, real=col, imag=row, helps with debugging
                            self.values[cnnz+k] = 0 #complex(ccol, self.row_idx[cnnz+k])

                        # keep track of how many nnz we have actually done...
                        cnnz += sm.rows
                        crow += sm.rows
                    elif sm.type == 'd': # a diagonal
                        self.row_idx[cnnz] = crow + j # set the diagonal position
                        if ccol == self.row_idx[cnnz]:
                            self.values[cnnz] = diagonal_fill
                        else:
                            self.values[cnnz] = complex(0, 0) #complex(ccol, self.row_idx[cnnz])
                        cnnz += 1
                        crow += 1
                    else:
                        raise Exception("Unhandled")

                # increment to the next column
                ccol += 1

        self.col_ptr[self.num_eqs] = cnnz
        self.__nnz = cnnz

        for cb in self.__callbacks:
            cb._updateview_()

    cdef np.ndarray get_matrix(self, SuiteSparse_long _from, SuiteSparse_long _to, complex_t** start_ptr, SuiteSparse_long* from_rhs_index):
        """
        Returns the submatrix that describes the coupling from a given
        block to another. For example, if you know the index for the block
        describing the HOM in a particular frequency, you can get the coupling
        submatrix to another set of HOMs at a frequency at a different node.

        This requires that `add_diagonal_elements` has been called to define a
        block with a certain index. Then `get_sub_matrix_view` must be called
        to state that a coupling will exist between two blocks.

        A block is a set of equations grouped together. A dense matrix decribes
        the coupling between blocks at different nodes. Blocks can have different
        shapes.

        .. todo:: What does `from_rhs_index` do?

        Parameters
        ----------
        _from : int
            Index of the block for coupling from
        _to : int
            Index of the block for coupling to
        start_ptr : complex_t**
            where to store memory start pointer
        """
        cdef:
            Py_ssize_t cdx = 0 # index where subcolumn starts in values array
            Py_ssize_t rowcount = 0 # number of non-zero rows in subcolumn
            Py_ssize_t rows
            Py_ssize_t i

            str _type = ""
            Py_ssize_t sdx = 0 # actual row index where submatrix starts
            complex_t[:] ptr

            _Column sub_col
            _SubMatrix sm

        from_rhs_index[0] = self.diag_map[_from]
        sub_col = self.sub_columns[_from]
        # actual index in sparse format where this subcolumn starts
        cdx = sub_col.start
        cols = sub_col.size

        # Loop over each submatrix in the subcolumn
        for sm in sub_col.submatrices:
            if sm.to == _to:
                _type = sm.type
                sdx   = rowcount
                rows  = sm.rows

            if sm.type == "d":
                rowcount += 1
            elif sm.type == "m":
                rowcount += sm.rows
            else:
                raise Exception("Unexpected result {}".format(sm.type))

        # get a memoryview of the entire subcolumn
        ptr = <complex_t[:(rowcount*cols)]>(self.values + cdx)

        # numpy array view of the matrix which we reshape into the proper
        # subcolumn size
        cdef np.ndarray arr = np.asarray(ptr).reshape(cols, rowcount).T
        cdef np.ndarray[complex, ndim=1] rtnD = None
        cdef np.ndarray[complex, ndim=2] rtnM = None

        # now we return a numpy view of just the part of the matrix requested
        if _type == "d":
            rtnD = arr[sdx, :]
        elif _type == "m":
            rtnM = arr[sdx:(sdx+rows), :]
        else:
            raise Exception("Submatrix connecting {} -> {} does not exist".format(_from, _to))

        if start_ptr != NULL:
            if rtnD is not None:
                start_ptr[0] = &rtnD[0]
            else:
                start_ptr[0] = &rtnM[0,0]

        if rtnD is None:
            return rtnM
        else:
            return rtnD

    cdef complex_t[::1] get_rhs_view(self, unsigned index):
        """
        Returns a view of the rhs vector corresponding to `index`.

        Parameters
        ----------
        index : unsigned
            The rhs vector to return a view of
        """
        if index >= self.num_rhs:
            raise ValueError(f"Invalid rhs index {index}")
        return self.rhs_view[index]

    @property
    def num_equations(self):
        """Returns the number of equations (rows) in this matrix."""
        return self.num_eqs

    def get_matrix_elements(self):
        """
        Returns the sparse CCS format for the current state of this matrix.

        Returns
        -------
        sparse : list(tuple())
            It is a list of (col, row, value)
        """
        vals = []
        ccol = -1
        for i in range(self.__nnz):
            if self.col_ptr[ccol+1] == i:
                ccol += 1
            vals.append((ccol, self.row_idx[i], self.values[i]))

        return vals

    def print_matrix(self):
        """Print a view of the non-zero elements in this matrix."""
        cidx = {}
        C = 0

        for col in self.sub_columns:
            for mat in self.sub_columns[col].submatrices:
                if mat.name[0] == 'I':
                    for i in range(mat.rows):
                        cidx[C] = f"{mat.name} mode={i}"
                        C+=1


        N = len(self.sub_columns)
        Ms = np.zeros((N,N),dtype=str)
        Ms[:] = ' '

        for _ in self.sub_columns:
            for __ in self.sub_columns[_].submatrices:
                Ms[__.to, self.sub_columns[_].index] = __.type

        #print(Ms)

        print("")
        print(f"Matrix {self.name}: nnz={self.__nnz} neqs={self.num_eqs}")
        print("    (col, row) = value")
        ccol = -1

        longest = np.zeros(2, dtype=np.int)
        for i in range(self.__nnz):
            if self.col_ptr[ccol+1] == i:
                ccol += 1
            longest[0] = max(longest[0], len(f"({ccol}, {self.row_idx[i]})"))
            longest[1] = max(longest[1], len(f"{self.values[i]}"))

        ccol = -1
        for i in range(self.__nnz):
            if self.col_ptr[ccol+1] == i:
                ccol += 1

            idx = f"({ccol}, {self.row_idx[i]})"
            value = f"{self.values[i]}"
            print(f"    {idx:{longest[0]}} = {value:{longest[1]}} : {cidx[ccol] if ccol in cidx else ''} -> {cidx[self.row_idx[i]] if self.row_idx[i] in cidx else ''}")

    def print_rhs(self, unsigned rhs_index=0):
        """
        Print a view of the rhs vector corresponding to `rhs_index`.

        Parameters
        ----------
        rhs_index : unsigned, optional
            The rhs vector to print; defaults to 0
        """
        cdef int i
        cdef const complex_t[::1] rhs
        cidx = {}
        C = 0

        if rhs_index >= self.num_rhs:
            raise ValueError(f"Invalid rhs index {rhs_index}")
        rhs = self.rhs_view[rhs_index]

        for col in self.sub_columns:
            for mat in self.sub_columns[col].submatrices:
                if mat.name[0] == 'I':
                    for i in range(mat.rows):
                        cidx[C] = f"{mat.name} mode={i}"
                        C+=1
        print("")
        print(f"Vector {self.name}: neqs={self.num_eqs}")
        print("    (row) = value")

        longest = np.zeros(2, dtype=np.int)
        for i in range(self.num_eqs):
            longest[0] = max(longest[0], len(f"({i})"))
            longest[1] = max(longest[1], len(f"{rhs[i]}"))

        for i in range(self.num_eqs):
            idx = f"({i})"
            value = f"{rhs[i]}"
            print(f"    {idx:{longest[0]}} = {value:{longest[1]}} : {cidx[i]}")


    cpdef clear_rhs(self, unsigned rhs_index=0):
        """
        Zero all elements in the rhs vector corresponding to `rhs_index`.

        Parameters
        ----------
        rhs_index : unsigned, optional
            The rhs vector to clear; defaults to 0
        """
        if rhs_index >= self.num_rhs:
            raise ValueError(f"Invalid rhs index {rhs_index}")
        memset(&self.rhs[rhs_index * self.num_eqs], 0, self.num_eqs*sizeof(complex_t))

    cpdef factor(self):
        raise NotImplementedError()

    cpdef refactor(self):
        raise NotImplementedError()

    cpdef const complex_t[::1] solve(self, int transpose=False, bint conjugate=False, unsigned rhs_index=0):
        raise NotImplementedError()

    cpdef void solve_extra_rhs(self, int transpose=False, bint conjugate=False):
        raise NotImplementedError()

    cdef void zgemv(self, complex_t[::1] out, unsigned rhs_index=0):
        raise NotImplementedError()

cdef class SubCCSView:
    def __init__(self, CCSMatrix Matrix, Py_ssize_t _from, Py_ssize_t _to, unicode name, bint conjugate_fill):
        self.M = weakref.ref(Matrix)
        self.A = None
        self._from  = _from
        self._to    = _to
        self.__shape = None
        self.__strides = None
        self.conjugate_fill = conjugate_fill

    @property
    def from_idx(self): return self._from

    @property
    def to_idx(self): return self._to

    @property
    def shape(self):
        return self.__shape

    @property
    def view(self):
        return self.A

    def __getitem__(self, key):
        return self.A[key]

    def __setitem__(self, key, value):
        # replace all elements
        if self.conjugate_fill:
            self.A[key] = -value.conjugate() # minus sign for off-diag blocks
        else:
            self.A[key] = -value # minus sign for off-diag blocks

    def _updateview_(self):
        # Here we store the numpy array wrapper for the memory location
        # as well as some details about it. The details are used in the
        # fast filling routines
        cdef CCSMatrix m = self.M()
        self.A = m.get_matrix(self._from, self._to, &self.ptr, &self.from_rhs_index)
        self.size1 = self.A.shape[0]
        self.stride1 = self.A.strides[0]//16
        if self.A.ndim == 2:
            self.size2 = self.A.shape[1]
            self.stride2 = self.A.strides[1]//16
            self.from_rhs_view = m.rhs_view[:, self.from_rhs_index:(self.from_rhs_index+self.size2)]
        else:
            self.size2 = 0
            self.stride2 = 0
            self.from_rhs_view = m.rhs_view[:, self.from_rhs_index:(self.from_rhs_index+self.size1)]

    cdef void fill_za(self, complex_t a):
        raise NotImplementedError()

    cdef void fill_zd(self, complex_t[::1] D):
        raise NotImplementedError()

    cdef void fill_dd(self, double[::1] D):
        raise NotImplementedError()

    cdef void fill_za_dd(self, complex_t a, double[::1] D):
        raise NotImplementedError()

    cdef void fill_zd_2(self, const complex_t* D, int s1) nogil:
        raise NotImplementedError()

    cdef void fill_za_zd_2(self, complex_t a, const complex_t* D, int stride) nogil:
        raise NotImplementedError()

    cdef void fill_za_zm(self, complex_t a, complex_t[:,::1] M):
        raise NotImplementedError()

    cdef void fill_za_zm_2(self, complex_t a, const complex_t* M, int s1, int s2):
        raise NotImplementedError()

    cdef void fill_za_zmc(self, complex_t a, const complex_t* M, int s1, int s2):
        raise NotImplementedError()

    cdef void fill_za_zmv(self, complex_t a, DenseZMatrix* M, DenseZVector* V):
        raise NotImplementedError()

    cdef void fill_za_zmvc(self, complex_t a, DenseZMatrix* M, DenseZVector* V):
        raise NotImplementedError()

    cdef void fill_zm(self, complex_t[:,::1] M):
        raise NotImplementedError()

    cdef void fill_negative_za(self, complex_t a):
        raise NotImplementedError()

    cdef void fill_negative_zd(self, complex_t[::1] D):
        raise NotImplementedError()

    cdef void fill_negative_dd(self, double[::1] D):
        raise NotImplementedError()

    cdef void fill_negative_za_dd(self, complex_t a, double[::1] D):
        raise NotImplementedError()

    cdef void fill_negative_zd_2(self, const complex_t* D, int s1) nogil:
        raise NotImplementedError()

    cdef void fill_negative_za_zd_2(self, complex_t a, const complex_t* D, int stride) nogil:
        raise NotImplementedError()

    cdef void fill_negative_za_zm(self, complex_t a, complex_t[:,::1] M):
        raise NotImplementedError()

    cdef void fill_negative_za_zm_2(self, complex_t a, const complex_t* M, int s1, int s2):
        raise NotImplementedError()

    cdef void fill_negative_za_zmc(self, complex_t a, const complex_t* M, int s1, int s2):
        raise NotImplementedError()

    cdef void fill_negative_za_zmv(self, complex_t a, DenseZMatrix* M, DenseZVector* V):
        raise NotImplementedError()

    cdef void fill_negative_za_zmvc(self, complex_t a, DenseZMatrix* M, DenseZVector* V):
        raise NotImplementedError()

    cdef void fill_negative_zm(self, complex_t[:,::1] M):
        raise NotImplementedError()

    cdef void fill_prop_za_zm(self, SubCCSView V, Py_ssize_t rhs_idx, complex_t a, complex_t[:,::1] M, bint increment):
        raise NotImplementedError()

    cdef void fill_neg_prop_za_zm(self, SubCCSView V, Py_ssize_t rhs_idx, complex_t a, complex_t[:,::1] M, bint increment):
        raise NotImplementedError()

    cdef void fill_prop_za(self, SubCCSView V, Py_ssize_t rhs_idx, complex_t a, bint increment):
        raise NotImplementedError()

    cdef void fill_neg_prop_za(self, SubCCSView V, Py_ssize_t rhs_idx, complex_t a, bint increment):
        raise NotImplementedError()


cdef class SubCCSMatrixView(SubCCSView):
    """
    This class represents a sub-matrix view of a CCS sparse matrix. This allows
    code to access and set values without worrying about the underlying
    sparse compression being used. Although so far this is just for CCS
    formats.

    This object will get a view of a n-by-m sub-matrix starting at index (i,j).
    The values of his matrix will be set initially to the coordinates.
    """
    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_za(self, complex_t a):
        cdef Py_ssize_t i, j
        if self.conjugate_fill:
            a = a.conjugate()

        for i in range(self.size1):
            for j in range(self.size2):
                self.ptr[i*self.stride1 + j*self.stride2] = a

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_zm(self, complex_t[:,::1] M):
        cdef Py_ssize_t i, j
        if self.conjugate_fill:
            for i in range(self.size1):
                for j in range(self.size2):
                    self.ptr[i*self.stride1 + j*self.stride2] = conj(M[i,j])
        else:
            for i in range(self.size1):
                for j in range(self.size2):
                    self.ptr[i*self.stride1 + j*self.stride2] = M[i,j]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_za_zm(self, complex_t a, complex_t[:,::1] M):
        cdef Py_ssize_t i, j
        if self.conjugate_fill:
            for i in range(self.size1):
                for j in range(self.size2):
                    self.ptr[i*self.stride1 + j*self.stride2] = conj(a * M[i,j])
        else:
            for i in range(self.size1):
                for j in range(self.size2):
                    self.ptr[i*self.stride1 + j*self.stride2] = a * M[i,j]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_za_zm_2(self, complex_t a, const complex_t* M, int s1, int s2):
        cdef int i, j
        if self.conjugate_fill:
            for i in range(self.size1):
                for j in range(self.size2):
                    self.ptr[i*self.stride1 + j*self.stride2] = conj(a * M[i*s1 + j*s2])
        else:
            for i in range(self.size1):
                for j in range(self.size2):
                    self.ptr[i*self.stride1 + j*self.stride2] = a * M[i*s1 + j*s2]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_za_zmc(self, complex_t a, const complex_t* M, int s1, int s2):
        cdef int i, j
        if self.conjugate_fill:
            a = conj(a)
            for i in range(self.size1):
                for j in range(self.size2):
                    self.ptr[i*self.stride1 + j*self.stride2] = a * M[i*s1 + j*s2]
        else:
            for i in range(self.size1):
                for j in range(self.size2):
                    self.ptr[i*self.stride1 + j*self.stride2] = a * conj(M[i*s1 + j*s2])

    def test_za_zm_2(self, complex a, np.ndarray[complex, ndim=2, mode='c'] M):
        self.fill_za_zm_2(a, <complex_t*>&M[0,0], M[:].strides[0]//16, M[:].strides[1]//16)

    # @cython.boundscheck(False)
    # @cython.wraparound(False)
    # @cython.initializedcheck(False)
    # cdef void fill_za_zmv(self, complex_t a, DenseZMatrix* M, DenseZVector* V):
    #     cdef int i, j, k
    #     if self.conjugate_fill:
    #         for i in range(self.size1): # row
    #             for j in range(self.size2): # col
    #                 self.ptr[i*self.stride1 + j*self.stride2] = 0 # reset for M@V
    #                 for k in range(self.size1): # row
    #                     self.ptr[i*self.stride1 + j*self.stride2] += M[i*s1 + k*s2] * V[k*sv]
    #                 self.ptr[i*self.stride1 + j*self.stride2] *= a
    #                 self.ptr[i*self.stride1 + j*self.stride2].imag *= -1 # conjugate
    #     else:
    #         for i in range(self.size1): # row
    #             for j in range(self.size2): # col
    #                 self.ptr[i*self.stride1 + j*self.stride2] = 0
    #                 for k in range(self.size1): # row
    #                     self.ptr[i*self.stride1 + j*self.stride2] += M[i*s1 + k*s2] * V[k*sv]
    #                 self.ptr[i*self.stride1 + j*self.stride2] *= a

    def do_fill_za_zmvc(self, complex_t a, complex_t[:,::1] M, complex_t[::1] V):
        cdef DenseZVector v
        cdef DenseZMatrix m

        m.ptr = &M[0,0]
        m.size1 = M.shape[0]
        m.size2 = M.shape[1]
        m.stride1 = M.strides[0]//16
        m.stride2 = M.strides[1]//16

        v.ptr = &V[0]
        v.size = V.shape[0]
        v.stride = V.strides[0]//16

        self.fill_za_zmvc(a, &m, &v)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_za_zmvc(self, complex_t a, DenseZMatrix* M, DenseZVector* V):
        """Sets view of submatrix to a * M @ V
        """
        cdef int i, k, stride
        assert(self.size1 == 1 or self.size2 == 1) # This view must be some 1D array either col or row
        assert(M.size1 == V.size) # Make sure size of matrix is correct for product
        assert( # check we vector is right size for this input matrix
            (self.size1 == 1 and self.size2 == V.size)
            or (self.size2 == 1 and self.size1 == V.size)
        )
        if V.size == self.size1:
            stride = self.stride1
        else:
            stride = self.stride2

        for i in range(V.size): # for each output element
            self.ptr[i*stride] = 0 # reset for M@V
            # do the matrix product
            for k in range(M.size2): # k-th col
                self.ptr[i*stride] += M.ptr[i*M.stride1 + k*M.stride2] * V.ptr[k*V.stride]
            self.ptr[i*stride] *= a

        if self.conjugate_fill:
            for i in range(V.size):
                # Can't use .imag to just select the imaginary part otherwise
                # cython injects python code... so do a bit of pointer math to get
                # double* access to imaginary part
                ((<double*>self.ptr) + 2*i*stride + 1)[0] *= -1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_prop_za_zm(self, SubCCSView V, Py_ssize_t rhs_idx, complex_t a, complex_t[:,::1] M, bint increment):
        # TODO ddb this can be optimised later once I know it works properly
        # For a given view of a matrix, propagate the current solved solution in it
        # then multiply it by some factor and apply some matrix operator to it
        V_rhs = np.atleast_2d(V.from_rhs_view[rhs_idx]).T
        if self.conjugate_fill:
            if increment:
                self.view[:] += np.conj(M @ (a * V.view @ V_rhs))
            else:
                self.view[:] = np.conj(M @ (a * V.view @ V_rhs))
        else:
            if increment:
                self.view[:] += M @ (a * V.view @ V_rhs)
            else:
                self.view[:] = M @ (a * V.view @ V_rhs)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_neg_prop_za_zm(self, SubCCSView V, Py_ssize_t rhs_idx, complex_t a, complex_t[:,::1] M, bint increment):
        self.fill_prop_za_zm(V, rhs_idx, -a, M, increment)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_prop_za(self, SubCCSView M, Py_ssize_t rhs_idx, complex_t a, bint increment):
        cdef complex_t* V = &M.from_rhs_view[rhs_idx][0]
        cdef Py_ssize_t N, stride, i, k
        # this current subview that we'll fill should be a row or column vector
        # we don't worry here about whether the matrix product is with a col or row
        # inuput vector though
        if self.size1 == self.size2 == 1:
            N = 1
            stride = max(self.stride1, self.stride2)
        else:
            # does a matrix vector product, so output must be a vector, or a 1x1
            assert(self.size1 == 1 ^ self.size2 == 1, "self size wrong") 
            if self.size1 == 1:
                N = self.size2
                stride = self.stride2
            else:
                N = self.size1
                stride = self.stride1
            assert(M.size1 == N, "M.side1 != N")

        # By the nature of how we get the rhs slice of the vector associated with the submatrix
        # V, it should always be the right size for a matrix product
        for i in range(N): # for each output element
            if not increment:
                self.ptr[i*stride] = 0 # reset for M@V
            # do the matrix product
            if self.conjugate_fill:
                for k in range(M.size2): # k-th col
                    self.ptr[i*stride] += conj(a * M.ptr[i*M.stride1 + k*M.stride2] * V[k]) # V stride is always 1
            else:
                for k in range(M.size2): # k-th col
                    self.ptr[i*stride] += a*M.ptr[i*M.stride1 + k*M.stride2] * V[k] # V stride is always 1

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_neg_prop_za(self, SubCCSView V, Py_ssize_t rhs_idx, complex_t a, bint increment):
        self.fill_prop_za(V, rhs_idx, -a, increment)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_negative_za(self, complex_t a):
        self.fill_za(-a)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_negative_zm(self, complex_t[:,::1] M):
        self.fill_za_zm(-1, M)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_negative_za_zm(self, complex_t a, complex_t[:,::1] M):
        self.fill_za_zm(-a, M)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_negative_za_zm_2(self, complex_t a, const complex_t* M, int s1, int s2):
        self.fill_za_zm_2(-a, M, s1, s2)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_negative_za_zmc(self, complex_t a, const complex_t* M, int s1, int s2):
        self.fill_za_zmc(-a, M, s1, s2)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_negative_za_zmv(self, complex_t a, DenseZMatrix* M, DenseZVector* V):
        self.fill_za_zmv(-a, M, V)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_negative_za_zmvc(self, complex_t a, DenseZMatrix* M, DenseZVector* V):
        self.fill_za_zmvc(-a, M, V)


cdef class SubCCSMatrixViewDiagonal(SubCCSView):
    """
    This class represents a sub-matrix view of a CCS sparse matrix. This allows
    code to access and set values without worrying about the underlying
    sparse compression being used. Although so far this is just for CCS
    formats.

    This object will get a view of a n-by-m sub-matrix starting at index (i,j).
    The values of his matrix will be set initially to the coordinates.
    """
    cdef void fill_za(self, complex_t a):
        cdef int i
        if self.conjugate_fill:
            a = a.conjugate()

        for i in range(self.size1):
            self.ptr[i*self.stride1] = a

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_zd(self, complex_t[::1] D):
        cdef int i
        if self.conjugate_fill:
            for i in range(self.size1):
                self.ptr[i*self.stride1] = conj(D[i])
        else:
            for i in range(self.size1):
                self.ptr[i*self.stride1] = D[i]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_dd(self, double[::1] D):
        cdef int i
        for i in range(self.size1):
            self.ptr[i*self.stride1] = D[i]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_za_dd(self, complex_t a, double[::1] D):
        cdef int i
        for i in range(self.size1):
            self.ptr[i*self.stride1] = a * D[i]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_zd_2(self, const complex_t* D, int s1) nogil:
        cdef int i
        if self.conjugate_fill:
            for i in range(self.size1):
                self.ptr[i*self.stride1] = conj(D[i*s1])
        else:
            for i in range(self.size1):
                self.ptr[i*self.stride1] = D[i*s1]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_za_zd_2(self, complex_t a, const complex_t* D, int D_stride) nogil:
        cdef int i
        if self.conjugate_fill:
            for i in range(self.size1):
                self.ptr[i*self.stride1] = conj(D[i*D_stride]*a)
        else:
            for i in range(self.size1):
                self.ptr[i*self.stride1] = D[i*D_stride]*a

    cdef void fill_negative_za(self, complex_t a):
        self.fill_za(-a)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_negative_zd(self, complex_t[::1] D):
        cdef int i
        if self.conjugate_fill:
            for i in range(self.size1):
                self.ptr[i*self.stride1] = conj(-D[i])
        else:
            for i in range(self.size1):
                self.ptr[i*self.stride1] = -D[i]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_negative_dd(self, double[::1] D):
        self.fill_za_dd(-1, D)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_negative_za_dd(self, complex_t a, double[::1] D):
        self.fill_za_dd(-a, D)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_negative_zd_2(self, const complex_t* D, int s1) nogil:
        cdef int i
        if self.conjugate_fill:
            for i in range(self.size1):
                self.ptr[i*self.stride1] = conj(-D[i*s1])
        else:
            for i in range(self.size1):
                self.ptr[i*self.stride1] = -D[i*s1]

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.initializedcheck(False)
    cdef void fill_negative_za_zd_2(self, complex_t a, const complex_t* D, int D_stride) nogil:
        self.fill_za_zd_2(-a, D, D_stride)

    # def test_za(self, value):
    #     self.fill_negative_za(value)

    # def test_zd(self, value):
    #     self.fill_negative_zd(value)s


cdef class SubCCSView1DArray:
    """
    This is a class for storing sub-matrix views for
    coupling directly to another single SubCCSView. It
    offers a 1D `PyObject*` array which can be iterated
    over in C for fast access matrix data without having
    to do reference inc/dec in fast loops.

    It can be accessed from Python for setting views,
    however it doesn't support slicing or wraparounds.

    Examples
    --------

    Cython access to views should be cast to a `SubCCSView`
    before using it:

    >>>> (<SubCCSView>arr.views[i]).fill()

    This should result in no python calls when checking the
    Cython analysis information.

    If you store this `SubCCSView` into a variable then a
    reference count will happen.
    """
    def __cinit__(self, Py_ssize_t size):
        self.size = size
        self.views = <PyObject**> calloc(size, sizeof(PyObject*))

    @property
    def dimensions(self):
        return 1

    def __getitem__(self, key):
        cdef int idx = <int?>key
        if idx < 0 or idx >= self.size:
            raise IndexError(f"Index must be {0} <= key < {self.size}")

        if self.views[idx] == NULL:
            return None
        else:
            return <object>self.views[idx]

    def __setitem__(self, key, value):
        cdef int idx = <int?>key
        if idx < 0 or idx >= self.size:
            raise IndexError(f"Index must be {0}<= key < {self.size}")
        if value is not None and not isinstance(value, SubCCSView):
            raise ValueError("Value is not a derivative of SubCCSView")

        if self.views[idx] != NULL:
            # Decrease ref for anything stored already
            Py_XDECREF(self.views[idx])

        cdef PyObject* ptr

        if value is None:
            ptr = NULL
        else:
            ptr = <PyObject*>value
            Py_XINCREF(ptr)

        self.views[idx] = ptr

    def __dealloc__(self):
        cdef int i
        for i in range(self.size):
            if self.views[i] == NULL:
                Py_XDECREF(self.views[i])
        free(self.views)


cdef class SubCCSView2DArray:
    """
    This is a class for storing sub-matrix views. It
    offers a 2D `PyObject**` array which can be iterated
    over in C for fast access matrix data without having
    to do reference inc/dec in fast loops.

    It can be accessed from Python for setting views,
    however it doesn't support slicing or wraparounds.

    Examples
    --------

    Cython access to views should be cast to a `SubCCSView`
    before using it:

    >>>> (<SubCCSView>arr.views[i][j]).fill()

    This should result in no python calls when checking the
    Cython analysis information.

    If you store this `SubCCSView` into a variable then a
    reference count will happen.
    """
    def __cinit__(self, Py_ssize_t rows, Py_ssize_t cols):
        cdef int i
        self.rows = rows
        self.cols = cols
        self.shape = (rows, cols)
        self.views = <PyObject***> calloc(rows, sizeof(PyObject**))
        for i in range(rows):
            self.views[i] = <PyObject**> calloc(cols, sizeof(PyObject*))

    @property
    def dimensions(self):
        return 2

    def __getitem__(self, key):
        cdef tuple idx = key
        if len(key) != 2:
            raise IndexError(f"Index must be 2D")
        if idx[0] < 0 or idx[0] >= self.rows:
            raise IndexError(f"Row index must be {0} <= key < {self.rows}")
        if idx[1] < 0 or idx[1] >= self.cols:
            raise IndexError(f"Column index must be {1} <= key < {self.cols}")

        if self.views[idx[0]][idx[1]] == NULL:
            return None
        else:
            return <object>self.views[idx[0]][idx[1]]

    def __setitem__(self, key, value):
        cdef tuple idx = key
        if len(key) != 2:
            raise IndexError(f"Index must be 2D")
        if idx[0] < 0 or idx[0] >= self.rows:
            raise IndexError(f"Row index must be 0 <= key:{idx[0]} < {self.rows}")
        if idx[1] < 0 or idx[1] >= self.cols:
            raise IndexError(f"Column index must be 0 <= key:{idx[1]} < {self.cols}")
        if value is not None and not isinstance(value, SubCCSView):
            raise ValueError("Value is not a derivative of SubCCSView")

        if self.views[idx[0]][idx[1]] != NULL:
            # Decrease ref for anything stored already
            Py_XDECREF(self.views[idx[0]][idx[1]])

        cdef PyObject* ptr

        if value is None:
            ptr = NULL
        else:
            ptr = <PyObject*>value
            Py_XINCREF(ptr)

        self.views[idx[0]][idx[1]] = ptr

    def __dealloc__(self):
        cdef int i
        cdef int j
        for i in range(self.rows):
            for j in range(self.cols):
                if self.views[i][j] == NULL:
                    Py_XDECREF(self.views[i][j])
            free(self.views[i])
        free(self.views)


cdef class KLUMatrix(CCSMatrix):
    """An object representation of a CCS matrix with methods to factor
    and solve the matrix via KLU.
    """
    def __cinit__(self, unicode name, int klu_ordering=0, int klu_scale=2, int klu_btf=1, int klu_maxwork=0):
        self.Symbolic = NULL
        self.Numeric  = NULL

        klu_l_defaults(&self.Common)

        self.Common.ordering = klu_ordering
        self.Common.scale    = klu_scale
        self.Common.btf      = klu_btf
        self.Common.maxwork  = klu_maxwork

    def __dealloc__(self):
        if self.Numeric:  klu_zl_free_numeric(&self.Numeric, &self.Common)
        if self.Symbolic: klu_l_free_symbolic(&self.Symbolic, &self.Common)

    cpdef factor(self):
        """Factor the matrix."""
        self.Symbolic = klu_l_analyze(self.num_eqs,
                                      self.col_ptr,
                                      self.row_idx,
                                      &self.Common)

        if self.Common.status != KLU_OK:
            raise Exception("An error occurred in KLU during analysis: STATUS={} ({})".format(self.Common.status, status_string(self.Common.status)))

        self.Numeric = klu_zl_factor(self.col_ptr,
                                     self.row_idx,
                                     <double*>self.values,
                                     self.Symbolic,
                                     &(self.Common))

        if self.Common.status != KLU_OK:
            raise Exception("An error occurred in KLU during factoring: STATUS={} ({})".format(self.Common.status, status_string(self.Common.status)))

        klu_zl_sort(self.Symbolic, self.Numeric, &(self.Common))

        if self.Common.status != KLU_OK:
            raise Exception("Sort failed. STATUS={} ({})".format(self.Common.status, status_string(self.Common.status)))

    cpdef refactor(self):
        """Re-factor the matrix."""
        klu_zl_refactor(self.col_ptr,
                        self.row_idx,
                         <double*>self.values,
                         self.Symbolic,
                         self.Numeric,
                         &(self.Common))

        if self.Common.status != KLU_OK:
            raise Exception("An error occurred in KLU during refactor: STATUS={} ({})".format(self.Common.status, status_string(self.Common.status)))



    cpdef const complex_t[::1] solve(self, int transpose=False, bint conjugate=False, unsigned rhs_index=0):
        """
        Solve the matrix with options for transposing and conjugating.

        If `transpose` is False, solves the linear system :math:`Ax = b` using the
        ``Symbolic`` and ``Numeric`` objects stored by this class.

        Otherwise, solves the linear system :math:`A^T x = b` or :math:`A^H x = b`. The
        `conjugate` option is zero for :math:`A^T x = b` or non-zero for :math:`A^H x = b`.

        Parameters
        ----------
        transpose : bool
            Flag determining whether to solve the transpose of the matrix.

        conjugate : bool
            Flag determining whether to solve :math:`A^T x =b` or :math:`A^H x = b`
            for the transposed linear system.

        rhs_index : unsigned, optional
            Which rhs vector to solve for. If unset, the default rhs vector is used.

        Returns
        -------
        out : np.ndarray
            The (negative) solution vector.
        """
        if rhs_index >= self.num_rhs:
            raise ValueError(f"Invalid rhs index {rhs_index}")

        if transpose:
            klu_zl_tsolve(self.Symbolic, self.Numeric, self.num_eqs, 1, <double*>&self.rhs[rhs_index * self.num_eqs], conjugate, &self.Common);
        else:
            klu_zl_solve( self.Symbolic, self.Numeric, self.num_eqs, 1, <double*>&self.rhs[rhs_index * self.num_eqs], &self.Common);

        # rtn = self.rhs_view.copy()
        # rtn.flags.writeable = False

        # return rtn
        return self.rhs_view[rhs_index]

    cpdef void solve_extra_rhs(self, int transpose=False, bint conjugate=False):
        """
        Solve the matrix for all present rhs vectors except the main one, with options for
        transposing and conjugating.

        If `transpose` is False, solves the linear system :math:`Ax = b` using the
        ``Symbolic`` and ``Numeric`` objects stored by this class.

        Otherwise, solves the linear system :math:`A^T x = b` or :math:`A^H x = b`. The
        `conjugate` option is zero for :math:`A^T x = b` or non-zero for :math:`A^H x = b`.

        As multiple rhs vectors are solved simultaneously, the result is not returned here,
        and must be retrieved via ``get_rhs_view``.

        Parameters
        ----------
        transpose : bool
            Flag determining whether to solve the transpose of the matrix.

        conjugate : bool
            Flag determining whether to solve :math:`A^T x =b` or :math:`A^H x = b`
            for the transposed linear system.
        """
        if transpose:
            klu_zl_tsolve(self.Symbolic, self.Numeric, self.num_eqs, self.num_rhs - 1, <double*>&self.rhs[self.num_eqs], conjugate, &self.Common);
        else:
            klu_zl_solve( self.Symbolic, self.Numeric, self.num_eqs, self.num_rhs - 1, <double*>&self.rhs[self.num_eqs], &self.Common);

    cpdef double rgrowth(self):
        """klu_rgrowth : compute the reciprocal pivot growth

        Pivot growth is computed after the input matrix is permuted, scaled, and
        off-diagonal entries pruned.  This is because the LU factorization of each
        block takes as input the scaled diagonal blocks of the BTF form.  The
        reciprocal pivot growth in column j of an LU factorization of a matrix C
        is the largest entry in C divided by the largest entry in U; then the overall
        reciprocal pivot growth is the smallest such value for all columns j.  Note
        that the off-diagonal entries are not scaled, since they do not take part in
        the LU factorization of the diagonal blocks.

        In MATLAB notation:
            rgrowth = min (max (abs ((R \ A(p,q)) - F)) ./ max (abs (U)))

        Returns
        -------
        reciprocal_pivot_growth : double
        """
        cdef SuiteSparse_long ok

        ok = klu_zl_rgrowth (
            self.col_ptr,
            self.row_idx,
            <double*>self.values,
            self.Symbolic,
            self.Numeric,
            &self.Common
        )

        if ok == 0:
            raise Exception("Error occurred whilst computing rgrowth")

        return self.Common.rgrowth

    cpdef double rcond(self):
        """ klu_rcond: compute min(abs(diag(U))) / max(abs(diag(U)))

        This function returns the smallest diagonal entry of U divided by the largest, which is a
        very crude estimate of the reciprocal of the condition number of the matrix A. It is very
        cheap to compute, however. In MATLAB notation, rcond = min(abs(diag(U))) /
        max(abs(diag(U))). If the matrix is singular, rcond will be zero.
        """
        cdef SuiteSparse_long ok

        ok = klu_zl_rcond (
            self.Symbolic,
            self.Numeric,
            &self.Common
        )

        if ok == 0:
            raise Exception("Error occurred whilst computing rcond")

        return self.Common.rcond

    cpdef double condest(self):
        """klu_condest

        Computes a reasonably accurate estimate of the 1-norm condition number, using
        Hager's method, as modified by Higham and Tisseur (same method as used in
        MATLAB's condest)
        """
        cdef SuiteSparse_long ok

        ok = klu_zl_condest (
            self.col_ptr,
            <double*>self.values,
            self.Symbolic,
            self.Numeric,
            &self.Common
        )

        if ok == 0:
            raise Exception("Error occurred whilst computing condest")

        return self.Common.condest

    cpdef void zgemv(self, complex_t[::1] out, unsigned rhs_index=0):
        """
        Multiply this matrix with the rhs vector corresponding to `rhs_index`, and store the result
        in `out`.

        Performs the operation :math:`y = A x`.

        Parameters
        ----------
        out : complex_t[::1]
            The vector to store the result in.
        rhs_index : unsigned, optional
            The rhs vector to multiply this matrix with; defaults to 0.
        """
        cdef:
            int ccol
            int crow

        if rhs_index >= self.num_rhs:
            raise ValueError(f"Invalid rhs index {rhs_index}")
        x = self.rhs_view[rhs_index]

        out[:] = 0
        ccol = -1
        for i in range(self.__nnz):
            if self.col_ptr[ccol+1] == i:
                ccol += 1

            crow = self.row_idx[i]

            out[ccol] += x[crow] * self.values[i]




import logging
LOGGER = logging.getLogger(__name__)

cimport cython
import finesse.constants
from finesse.cmatrix import SubCCSView
from finesse.cmatrix cimport SubCCSView
from finesse.components.workspace cimport ConnectorWorkspace, FillFuncWrapper
from finesse.components.workspace import Connections
from finesse.cymath cimport complex_t
from finesse.components import Connector, NodeType, NodeDirection
from finesse.parameter import float_parameter
from finesse.components.general import DOFDefinition

from cpython.ref cimport PyObject, Py_XINCREF, Py_XDECREF
from libc.string cimport strcmp, memcpy
from libc.stdlib cimport free, calloc

cimport numpy as np
import numpy as np

cdef extern from "constants.h":
    long double PI


def get_mechanical_port(connect_to):
    # Handle different types of elements or mech ports to connect to
    if isinstance(connect_to, Connector):
        mech_ports = [p for p in connect_to.ports if p.type == NodeType.MECHANICAL]
        if len(mech_ports) > 1:
            raise Exception(f"{connect_to} has more than one mechanical node so please specify which to use.")
        return mech_ports[0]
    else:
        return connect_to


cdef class MIMOTFWorkspace(ConnectorWorkspace):
    """Workspace that contains MIMO transfer functions stored
    in a numerator/denominator basis.
    """

    cdef:
        double[::1] denom
        Py_ssize_t N_num_allocd
        Py_ssize_t curr_num
        double** numerators
        int* numerator_sizes
        complex_t curr_denom
        complex_t s

    def __cinit__(self, owner, sim, refill, unsigned int N_numerators):
        self.N_num_allocd = N_numerators
        self.numerators = <double**>calloc(N_numerators, sizeof(double*))
        if not self.numerators:
            raise MemoryError()
        self.numerator_sizes = <int*>calloc(N_numerators, sizeof(int))
        if not self.numerator_sizes:
            raise MemoryError()
        self.curr_num = 0
        self.curr_denom = 0

    def __init__(self, owner, sim, bint refill, unsigned int N_numerators):
        super().__init__(
            owner,
            sim,
            refill,
            refill,
            Connections(),
            Connections()
        )
        self.signal.add_fill_function(mimo_fill, refill)

    def __dealloc__(self):
        if self.numerators:
            free(self.numerators)
        if self.numerator_sizes:
            free(self.numerator_sizes)

    @property
    def num_numerators(self):
        return self.curr_num

    def set_denominator(self, double[::1] denom):
        self.denom = denom

    def add_numerator(self, double[::1] num):
        if self.curr_num == self.N_num_allocd:
            raise Exception("Added more numerators than were allocated for.")
        self.numerators[self.curr_num] = &num[0]
        self.numerator_sizes[self.curr_num] = len(num)
        self.curr_num += 1

    cpdef void set_s(self, complex_t s):
        self.s = s
        self.curr_denom = eval_tf_term(s, &self.denom[0], len(self.denom))

    cpdef complex_t H(self, int numerator_idx):
        if not (0 <= numerator_idx < self.curr_num):
            raise Exception("Unexpected index")

        return eval_tf(
            self.s,
            self.numerators[numerator_idx],
            self.numerator_sizes[numerator_idx],
            self.curr_denom
        )


@cython.wraparound(False)
@cython.boundscheck(False)
@cython.initializedcheck(False)
cdef inline complex_t eval_tf_term(complex_t s, const double* coeffs, int N):
    cdef:
        int i
        complex res = 0
    for i in range(N):
        res = res * s + coeffs[i]

    return res


@cython.wraparound(False)
@cython.boundscheck(False)
@cython.initializedcheck(False)
cdef inline complex_t eval_tf(complex_t s, const double* num, int N, complex_t den):
    return eval_tf_term(s, num, N)/den


@cython.wraparound(False)
@cython.boundscheck(False)
@cython.initializedcheck(False)
cpdef eval_tf_vec(const complex_t[::1] s, const double[::1] num, const double[::1] den, complex_t[::1] out):
    cdef:
        int i
        complex_t oden = 0
        int N = len(s)
        int Nn = len(num)
        int Nd = len(den)

    if len(out) != len(s):
        raise Exception("Length of `s` differs from output `out`")

    for i in range(N):
        out[i] = eval_tf_term(s[i], &num[0], Nn)/eval_tf_term(s[i], &den[0], Nd)

    return 0

mimo_fill = FillFuncWrapper.make_from_ptr(c_mimo_fill)
@cython.wraparound(False)
@cython.boundscheck(False)
@cython.initializedcheck(False)
cdef c_mimo_fill(ConnectorWorkspace cws):
    cdef MIMOTFWorkspace ws = <MIMOTFWorkspace>cws
    cdef complex_t s = 0
    cdef tuple key
    s.imag = 2 * PI * ws.sim.model_data.fsig
    # Sets the complex s value for this step and precomputes the denominator
    ws.set_s(s)
    for i in range(ws.curr_num):
        key = (ws.owner_id,i,0,0)
        if key in ws.sim.signal._submatrices:
            (<SubCCSView>ws.sim.signal._submatrices[key]).fill_negative_za(ws.H(i))


class Joint(Connector):
    """Represents a mechanical joint between two mechanical ports in the model.

    TODO:
        - Need some way of specifying transfer functions for coupling
          between the two ports in question.

    Parameters
    ----------
    name : str, optional
        Name of newly created joint.

    portA, portB : :class:`.Port`, optional
        Ports to connect.

    connections : ?
    """

    def __init__(self, name=None, portA=None, portB=None, connections=None):
        if (portA is None) != (portB is None):
            LOGGER.warn(
                "Can't construct a joint with only one port "
                "connected, ignoring ports."
            )
            portA = None
            portB = None

        if portA is not None and portA.type != NodeType.MECHANICAL:
            raise Exception("PortA argument is not a mechanical port")
        if portB is not None and portB.type != NodeType.MECHANICAL:
            raise Exception("PortB argument is not a mechanical port")

        if name is None:
            if portA is not None and portB is not None:
                compA = portA.component.name
                compB = portB.component.name
                name = f"{compA}_{portA.name}__{compB}_{portB.name}"
            else:
                raise ValueError(
                    "Cannot create an unconnected wire without " "providing a name"
                )

        super().__init__(name)
        self._add_to_model_namespace = False

        self.portA = portA
        self.portB = portB

        """
        The mechanical ports have nodes that are the motion degrees of
        freedom, so there can potentially be many, but the usual are z,
        yaw, pitch. We just make copies of the nodes here that are in
        each mechanical port.
        """
        self._add_port("n1", NodeType.MECHANICAL)
        self._add_port("n2", NodeType.MECHANICAL)

        if portA is not None and portB is not None:
            self.connect(portA, portB)

    def connect(self, portA, portB):
        """
        Sets the ports of this `Joint`.

        Parameters
        ----------
        portA : :class:`.Port`, optional
            Port to connect

        portB : :class:`.Port`, optional
            Port to connect
        """
        for n in portA.nodes:
            self.n1._add_node(n.name, None, n)

        for n in portB.nodes:
            self.n2._add_node(n.name, None, n)
        """
        I'm not sure of the perfect way to link up two mechanical ports.
        For now I'm going to link up matching names. So z motion of one
        mech port is linked to the z node of the other.

        If there is only one node in a port in both then just connect
        them up.
        """
        if len(portA.nodes) == len(portB.nodes) == 1:
            self._register_node_coupling(portA.nodes[0], portB.nodes[0])
        else:
            for na in portA.nodes:
                for nb in portB.nodes:
                    if na.name == nb.name:
                        self._register_node_coupling(na, nb)

    def _on_init(self, sim):
        # TODO: (sjr) does this need to have rows > 1 for HOM simulations?
        self.__I = np.eye(sim.model_data.num_HOMs, dtype=np.complex128)

    def _fill_matrix(self, sim):
        for c in list(self._registered_connections):
            with sim.signal.component_edge_fill(self, c, None, None) as mat:
                mat[:] = self.__I


class FreeMassWorkspace(ConnectorWorkspace):
    pass


@float_parameter("mass", "Mass", units="kg")
@float_parameter("I_pitch", "Moment of inertia (pitch)", units="kg·m^2")
@float_parameter("I_yaw", "Moment of inertia (yaw)", units="kg·m^2")
class FreeMass(Connector):
    """Simple free mass suspension of an object.

    The object being suspended must have a mechanical port with
    nodes z, pitch, and yaw and forces F_z, F_pitch, and F_yaw.
    """

    def __init__(self, name, connect_to, mass=np.inf, I_yaw=np.inf, I_pitch=np.inf):
        super().__init__(name)
        mech_port = get_mechanical_port(connect_to)
        self.mass = mass
        self.I_yaw = I_yaw
        self.I_pitch = I_pitch

        # Add motion and force nodes to mech port.
        # Here we duplicate the already created mechanical
        # nodes in some other connector element
        self._add_port("mech", NodeType.MECHANICAL)
        self.mech._add_node("z", None, mech_port.z)
        self.mech._add_node("yaw", None, mech_port.yaw)
        self.mech._add_node("pitch", None, mech_port.pitch)
        self.mech._add_node("F_z", None, mech_port.F_z)
        self.mech._add_node("F_yaw", None, mech_port.F_yaw)
        self.mech._add_node("F_pitch", None, mech_port.F_pitch)
        # We just have direct coupling between DOF, no cross-couplings
        self._register_node_coupling(
            "F_to_Z", self.mech.F_z, self.mech.z,
            enabled_check=lambda: float(self.mass) < np.inf and not self.mass.is_changing
        )
        self._register_node_coupling(
            "F_to_YAW", self.mech.F_yaw, self.mech.yaw,
            enabled_check=lambda: float(self.I_yaw) < np.inf and not self.I_yaw.is_changing
        )
        self._register_node_coupling(
            "F_to_PITCH", self.mech.F_pitch, self.mech.pitch,
            enabled_check=lambda: float(self.I_pitch) < np.inf and not self.I_pitch.is_changing
        )
        # Define typical degrees of freedom for this component
        import types
        self.dofs = types.SimpleNamespace()
        self.dofs.z = DOFDefinition(None, self.mech.z, 1)
        self.dofs.F_z = DOFDefinition(None, self.mech.F_z, 1)

    def _get_workspace(self, sim):
        if sim.signal:
            refill = sim.model.fsig.f.is_changing or any(p.is_changing for p in self.parameters)
            ws = FreeMassWorkspace(self, sim, refill, refill)
            ws.signal.add_fill_function(self.fill, refill)
            return ws
        else:
            return None

    def fill(self, ws):
        f = ws.sim.model_data.fsig
        if ws.signal.connections.F_to_Z_idx >= 0:
            with ws.sim.signal.component_edge_fill3(
                ws.owner_id, ws.signal.connections.F_to_Z_idx, 0, 0,
            ) as mat:
                mat[:] = -1 / (ws.values.mass * (2*PI*f)**2)

        if ws.signal.connections.F_to_YAW_idx >= 0:
            with ws.sim.signal.component_edge_fill3(
                ws.owner_id, ws.signal.connections.F_to_YAW_idx, 0, 0,
            ) as mat:
                mat[:] = -1 / (ws.values.I_yaw * (2*PI*f)**2)

        if ws.signal.connections.F_to_PITCH_idx >= 0:
            with ws.sim.signal.component_edge_fill3(
                ws.owner_id, ws.signal.connections.F_to_PITCH_idx, 0, 0,
            ) as mat:
                mat[:] = -1 / (ws.values.I_pitch * (2*PI*f)**2)


class PendulumMassWorkspace(ConnectorWorkspace):
    pass


@float_parameter("mass", "Mass", units="kg")
@float_parameter("Qz", "Qz", units="")
@float_parameter("fz", "fz", units="Hz")
@float_parameter("I_pitch", "Moment of inertia (pitch)", units="kg·m^2")
@float_parameter("Qyaw", "Qyaw", units="")
@float_parameter("fyaw", "fyaw", units="Hz")
@float_parameter("I_yaw", "Moment of inertia (yaw)", units="kg·m^2")
@float_parameter("Qpitch", "Qpitch", units="")
@float_parameter("fpitch", "fpitch", units="Hz")
class Pendulum(Connector):
    """Simple pendulum suspension of an object.

    The object being suspended must have a mechanical port with
    nodes z, pitch, and yaw and forces F_z, F_pitch, and F_yaw.
    """

    def __init__(self, name, mech_port, mass=np.inf, Qz=1000, fz=1, I_yaw=np.inf, Qyaw=1000, fyaw=1, I_pitch=np.inf, Qpitch=1000, fpitch=1):
        super().__init__(name)
        self.mass = mass
        self.Qz = Qz
        self.fz = fz
        self.Qyaw = Qyaw
        self.fyaw = fyaw
        self.Qpitch = Qpitch
        self.fpitch = fpitch
        self.I_yaw = I_yaw
        self.I_pitch = I_pitch

        # Add motion and force nodes to mech port.
        # Here we duplicate the already created mechanical
        # nodes in some other connector element
        self._add_port("mech", NodeType.MECHANICAL)
        self.mech._add_node("z", None, mech_port.z)
        self.mech._add_node("yaw", None, mech_port.yaw)
        self.mech._add_node("pitch", None, mech_port.pitch)
        self.mech._add_node("F_z", None, mech_port.F_z)
        self.mech._add_node("F_yaw", None, mech_port.F_yaw)
        self.mech._add_node("F_pitch", None, mech_port.F_pitch)
        # We just have direct coupling between DOF, no cross-couplings
        self._register_node_coupling(
            "F_to_Z", self.mech.F_z, self.mech.z,
            enabled_check=lambda: float(self.mass) < np.inf and not self.mass.is_changing
        )
        self._register_node_coupling(
            "F_to_YAW", self.mech.F_yaw, self.mech.yaw,
            enabled_check=lambda: float(self.I_yaw) < np.inf and not self.I_yaw.is_changing
        )
        self._register_node_coupling(
            "F_to_PITCH", self.mech.F_pitch, self.mech.pitch,
            enabled_check=lambda: float(self.I_pitch) < np.inf and not self.I_pitch.is_changing
        )

    def _get_workspace(self, sim):
        if sim.signal:
            refill = sim.model.fsig.f.is_changing or any(p.is_changing for p in self.parameters)
            ws = PendulumMassWorkspace(self, sim, refill, refill)
            ws.signal.add_fill_function(self.fill, refill)
            return ws
        else:
            return None

    def fill(self, ws):
        cdef:
            complex_t s = 2j* PI * ws.sim.model_data.fsig
            complex_t num = 1
            complex_t den = 1
            double omega0

        if ws.signal.connections.F_to_Z_idx >= 0:
            with ws.sim.signal.component_edge_fill3(
                ws.owner_id, ws.signal.connections.F_to_Z_idx, 0, 0,
            ) as mat:
                omega0 = 2 * PI * ws.values.fz
                mat[:] = 1 / ws.values.mass * 1/(s**2  + s * omega0/ws.values.Qz + omega0**2)

        if ws.signal.connections.F_to_YAW_idx >= 0:
            with ws.sim.signal.component_edge_fill3(
                ws.owner_id, ws.signal.connections.F_to_YAW_idx, 0, 0,
            ) as mat:
                omega0 = 2 * PI * ws.values.fyaw
                mat[:] = 1 / ws.values.I_yaw * 1/(s**2  + s * omega0/ws.values.Qyaw + omega0**2)

        if ws.signal.connections.F_to_PITCH_idx >= 0:
            with ws.sim.signal.component_edge_fill3(
                ws.owner_id, ws.signal.connections.F_to_PITCH_idx, 0, 0,
            ) as mat:
                omega0 = 2 * PI * ws.values.fpitch
                mat[:] = 1 / ws.values.I_pitch * 1/(s**2  + s * omega0/ws.values.Qpitch + omega0**2)



cimport numpy as np
from finesse.cymath cimport complex_t
from finesse.cymath.gaussbeam cimport beam_param
from finesse.tracing.ctracer cimport TraceForest, TraceTree
from finesse.frequency cimport frequency_info_t, FrequencyContainer
from cpython.ref cimport PyObject


cdef struct NodeInfoEntry:
    Py_ssize_t index
    Py_ssize_t rhs_index
    Py_ssize_t freq_index
    Py_ssize_t nfreqs # number of frequencies
    Py_ssize_t nhoms # number of HOMs
    frequency_info_t *frequencies # Frequencies present at this node, size nfreqs


cdef struct NodeBeamParam:
    beam_param qx
    beam_param qy
    bint is_changing


cdef class ModelData:
    cdef public:
        double fsig
    cdef readonly:
        double EPSILON0_C, UNIT_VACUUM, lambda0
        double f0, k0
        int num_HOMs
        int[:, ::1] homs_view
        int maxtem
        int max_n # Maximum mode index in tangential plane
        int max_m # Maximum mode index in sagittal plane
        double x_scale
        bint zero_K00 # should phase of k0000 coefficients be zeroed
        bint zero_tem00_gouy # should Gouy phase of TEM00 be zeroed
        bint v2_transmission_phase
    cdef:
        # Flat contiguous ptr to homs_view
        int* homs_ptr


# Contains non-physical configuration data, as low-level values, to be
# used during a simulation
cdef struct SimConfigData:
    # Number of threads to use for prange loops whose
    # size is proportional to the number of modes
    int nthreads_homs
